/* original parser id follows */
/* yysccsid[] = "@(#)yaccpar	1.9 (Berkeley) 02/21/93" */
/* (use YYMAJOR/YYMINOR for ifdefs dependent on parser version) */

#define YYBYACC 1
#define YYMAJOR 1
#define YYMINOR 9
#define YYPATCH 20150711

#define YYEMPTY        (-1)
#define yyclearin      (yychar = YYEMPTY)
#define yyerrok        (yyerrflag = 0)
#define YYRECOVERING() (yyerrflag != 0)
#define YYENOMEM       (-2)
#define YYEOF          0
#define YYPREFIX "yy"

#define YYPURE 0

#line 2 "src/asm/asmy.y"
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "asm/symbol.h"
#include "asm/asm.h"
#include "asm/charmap.h"
#include "asm/output.h"
#include "asm/mylink.h"
#include "asm/fstack.h"
#include "asm/mymath.h"
#include "asm/rpn.h"
#include "asm/main.h"
#include "asm/lexer.h"

char	*tzNewMacro;
ULONG	ulNewMacroSize;

void
bankrangecheck(char *name, ULONG secttype, SLONG org, SLONG bank)
{
	SLONG minbank, maxbank;
	char *stype = NULL;
	switch (secttype) {
	case SECT_ROMX:
		stype = "ROMX";
		minbank = 1;
		maxbank = 0x1ff;
		break;
	case SECT_SRAM:
		stype = "SRAM";
		minbank = 0;
		maxbank = 0x1ff;
		break;
	case SECT_WRAMX:
		stype = "WRAMX";
		minbank = 1;
		maxbank = 7;
		break;
	case SECT_VRAM:
		stype = "VRAM";
		minbank = 0;
		maxbank = 1;
		break;
	default:
		yyerror("BANK only allowed for "
		    "ROMX, WRAMX, SRAM, or VRAM sections");
	}

	if (stype && (bank < minbank || bank > maxbank)) {
		yyerror("%s bank value $%x out of range ($%x to $%x)",
		    stype, bank, minbank, maxbank);
	}

	if (secttype == SECT_WRAMX) {
		bank -= minbank;
	}

	out_NewAbsSection(name, secttype, org, bank);
}

size_t symvaluetostring(char *dest, size_t maxLength, char *sym)
{
	size_t length;

	if (sym_isString(sym)) {
		char *src = sym_GetStringValue(sym);
		size_t i;

		for (i = 0; src[i] != 0; i++) {
			if (i >= maxLength) {
				fatalerror("Symbol value too long to fit buffer");
			}
			dest[i] = src[i];
		}

		length = i;
	} else {
		ULONG value = sym_GetConstantValue(sym);
		int fullLength = snprintf(dest, maxLength + 1, "$%lX", value);

		if (fullLength < 0) {
			fatalerror("snprintf encoding error");
		} else {
			length = (size_t)fullLength;

			if (length > maxLength) {
				fatalerror("Symbol value too long to fit buffer");
			}
		}
	}

	return length;
}

ULONG	str2int( char *s )
{
	ULONG r=0;
	while( *s )
	{
		r<<=8;
		r|=(UBYTE)(*s++);
	}
	return( r );
}

ULONG	str2int2( char *s, int length )
{
	int i;
	ULONG r=0;
	i = (length - 4 < 0 ? 0 : length - 4);
	while(i < length)
	{
		r<<=8;
		r|=(UBYTE)(s[i]);
		i++;
		
	}
	return( r );
}

ULONG	isWhiteSpace( char s )
{
	return( s==' ' || s=='\t' || s=='\0' || s=='\n' );
}

ULONG	isRept( char *s )
{
	return( (strncasecmp(s,"REPT",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

ULONG	isEndr( char *s )
{
	return( (strncasecmp(s,"Endr",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

void	copyrept( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;
	char	*bufferEnd = pCurrentBuffer->pBufferStart + pCurrentBuffer->nBufferSize;

	while( src < bufferEnd && level )
	{
		if( instring==0 )
		{
			if( isRept(src) )
			{
				level+=1;
				src+=4;
			}
			else if( isEndr(src) )
			{
				level-=1;
				src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	if (level != 0) {
		fatalerror("Unterminated REPT block");
	}

	len=src-pCurrentBuffer->pBuffer-4;

	src=pCurrentBuffer->pBuffer;
	ulNewMacroSize=len;

	if ((tzNewMacro = malloc(ulNewMacroSize + 1)) != NULL) {
		ULONG i;

		tzNewMacro[ulNewMacroSize]=0;
		for( i=0; i<ulNewMacroSize; i+=1 )
		{
			if( (tzNewMacro[i]=src[i])=='\n' )
				nLineNo+=1;
		}
	} else
		fatalerror( "No mem for REPT block" );

	yyskipbytes( ulNewMacroSize+4 );

}

ULONG	isMacro( char *s )
{
	return( (strncasecmp(s,"MACRO",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[5]) );
}

ULONG	isEndm( char *s )
{
	return( (strncasecmp(s,"Endm",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

void	copymacro( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;
	char	*bufferEnd = pCurrentBuffer->pBufferStart + pCurrentBuffer->nBufferSize;

	while( src < bufferEnd && level )
	{
		if( instring==0 )
		{
			if( isMacro(src) )
			{
				level+=1;
				src+=4;
			}
			else if( isEndm(src) )
			{
				level-=1;
				src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	if (level != 0) {
		fatalerror("Unterminated MACRO definition");
	}

	len=src-pCurrentBuffer->pBuffer-4;

	src=pCurrentBuffer->pBuffer;
	ulNewMacroSize=len;

	if( (tzNewMacro=(char *)malloc(ulNewMacroSize+2))!=NULL )
	{
		ULONG i;

		tzNewMacro[ulNewMacroSize]='\n';
		tzNewMacro[ulNewMacroSize+1]=0;
		for( i=0; i<ulNewMacroSize; i+=1 )
		{
			if( (tzNewMacro[i]=src[i])=='\n' )
				nLineNo+=1;
		}
	}
	else
		fatalerror( "No mem for MACRO definition" );

	yyskipbytes( ulNewMacroSize+4 );
}

ULONG	isIf( char *s )
{
	return( (strncasecmp(s,"If",2)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[2]) );
}

ULONG	isElse( char *s )
{
	return( (strncasecmp(s,"Else",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

ULONG	isEndc( char *s )
{
	return( (strncasecmp(s,"Endc",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

void	if_skip_to_else( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;

	while( *src && level )
	{
		if( *src=='\n' )
			nLineNo+=1;

		if( instring==0 )
		{
			if( isIf(src) )
			{
				level+=1;
				src+=2;
			}
			else if( level==1 && isElse(src) )
			{
				level-=1;
				src+=4;
			}
			else if( isEndc(src) )
			{
				level-=1;
				if( level!=0 )
					src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	if (level != 0) {
		fatalerror("Unterminated IF construct");
	}

	len=src-pCurrentBuffer->pBuffer;

	yyskipbytes( len );
	yyunput( '\n' );
	nLineNo-=1;
}

void	if_skip_to_endc( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;

	while( *src && level )
	{
		if( *src=='\n' )
			nLineNo+=1;

		if( instring==0 )
		{
			if( isIf(src) )
			{
				level+=1;
				src+=2;
			}
			else if( isEndc(src) )
			{
				level-=1;
				if( level!=0 )
					src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	if (level != 0) {
		fatalerror("Unterminated IF construct");
	}

	len=src-pCurrentBuffer->pBuffer;

	yyskipbytes( len );
	yyunput( '\n' );
	nLineNo-=1;
}

#line 433 "src/asm/asmy.y"
#ifdef YYSTYPE
#undef  YYSTYPE_IS_DECLARED
#define YYSTYPE_IS_DECLARED 1
#endif
#ifndef YYSTYPE_IS_DECLARED
#define YYSTYPE_IS_DECLARED 1
typedef union
{
    char tzSym[MAXSYMLEN + 1];
    char tzString[MAXSTRLEN + 1];
    struct Expression sVal;
    SLONG nConstValue;
} YYSTYPE;
#endif /* !YYSTYPE_IS_DECLARED */
#line 466 "src/asm/asmy.c"

/* compatibility with bison */
#ifdef YYPARSE_PARAM
/* compatibility with FreeBSD */
# ifdef YYPARSE_PARAM_TYPE
#  define YYPARSE_DECL() yyparse(YYPARSE_PARAM_TYPE YYPARSE_PARAM)
# else
#  define YYPARSE_DECL() yyparse(void *YYPARSE_PARAM)
# endif
#else
# define YYPARSE_DECL() yyparse(void)
#endif

/* Parameters sent to lex. */
#ifdef YYLEX_PARAM
# define YYLEX_DECL() yylex(void *YYLEX_PARAM)
# define YYLEX yylex(YYLEX_PARAM)
#else
# define YYLEX_DECL() yylex(void)
# define YYLEX yylex()
#endif

/* Parameters sent to yyerror. */
#ifndef YYERROR_DECL
#define YYERROR_DECL() yyerror(const char *s)
#endif
#ifndef YYERROR_CALL
#define YYERROR_CALL(msg) yyerror(msg)
#endif

extern int YYPARSE_DECL();

#define T_NUMBER 257
#define T_STRING 258
#define T_OP_LOGICNOT 259
#define T_OP_LOGICOR 260
#define T_OP_LOGICAND 261
#define T_OP_LOGICEQU 262
#define T_OP_LOGICGT 263
#define T_OP_LOGICLT 264
#define T_OP_LOGICGE 265
#define T_OP_LOGICLE 266
#define T_OP_LOGICNE 267
#define T_OP_ADD 268
#define T_OP_SUB 269
#define T_OP_OR 270
#define T_OP_XOR 271
#define T_OP_AND 272
#define T_OP_SHL 273
#define T_OP_SHR 274
#define T_OP_MUL 275
#define T_OP_DIV 276
#define T_OP_MOD 277
#define T_OP_NOT 278
#define T_OP_DEF 279
#define T_OP_BANK 280
#define T_OP_SIN 281
#define T_OP_COS 282
#define T_OP_TAN 283
#define T_OP_ASIN 284
#define T_OP_ACOS 285
#define T_OP_ATAN 286
#define T_OP_ATAN2 287
#define T_OP_FDIV 288
#define T_OP_FMUL 289
#define T_OP_ROUND 290
#define T_OP_CEIL 291
#define T_OP_FLOOR 292
#define T_OP_STRCMP 293
#define T_OP_STRIN 294
#define T_OP_STRSUB 295
#define T_OP_STRLEN 296
#define T_OP_STRCAT 297
#define T_OP_STRUPR 298
#define T_OP_STRLWR 299
#define NEG 300
#define T_LABEL 301
#define T_ID 302
#define T_POP_EQU 303
#define T_POP_SET 304
#define T_POP_EQUS 305
#define T_POP_INCLUDE 306
#define T_POP_PRINTF 307
#define T_POP_PRINTT 308
#define T_POP_PRINTV 309
#define T_POP_IF 310
#define T_POP_ELSE 311
#define T_POP_ENDC 312
#define T_POP_IMPORT 313
#define T_POP_EXPORT 314
#define T_POP_GLOBAL 315
#define T_POP_DB 316
#define T_POP_DS 317
#define T_POP_DW 318
#define T_POP_DL 319
#define T_POP_SECTION 320
#define T_POP_RB 321
#define T_POP_RW 322
#define T_POP_RL 323
#define T_POP_MACRO 324
#define T_POP_ENDM 325
#define T_POP_RSRESET 326
#define T_POP_RSSET 327
#define T_POP_INCBIN 328
#define T_POP_REPT 329
#define T_POP_CHARMAP 330
#define T_POP_SHIFT 331
#define T_POP_ENDR 332
#define T_POP_FAIL 333
#define T_POP_WARN 334
#define T_POP_PURGE 335
#define T_POP_POPS 336
#define T_POP_PUSHS 337
#define T_POP_POPO 338
#define T_POP_PUSHO 339
#define T_POP_OPT 340
#define T_SECT_WRAM0 341
#define T_SECT_VRAM 342
#define T_SECT_ROMX 343
#define T_SECT_ROM0 344
#define T_SECT_HRAM 345
#define T_SECT_WRAMX 346
#define T_SECT_SRAM 347
#define T_Z80_ADC 348
#define T_Z80_ADD 349
#define T_Z80_AND 350
#define T_Z80_BIT 351
#define T_Z80_CALL 352
#define T_Z80_CCF 353
#define T_Z80_CP 354
#define T_Z80_CPL 355
#define T_Z80_DAA 356
#define T_Z80_DEC 357
#define T_Z80_DI 358
#define T_Z80_EI 359
#define T_Z80_EX 360
#define T_Z80_HALT 361
#define T_Z80_INC 362
#define T_Z80_JP 363
#define T_Z80_JR 364
#define T_Z80_LD 365
#define T_Z80_LDI 366
#define T_Z80_LDD 367
#define T_Z80_LDIO 368
#define T_Z80_NOP 369
#define T_Z80_OR 370
#define T_Z80_POP 371
#define T_Z80_PUSH 372
#define T_Z80_RES 373
#define T_Z80_RET 374
#define T_Z80_RETI 375
#define T_Z80_RST 376
#define T_Z80_RL 377
#define T_Z80_RLA 378
#define T_Z80_RLC 379
#define T_Z80_RLCA 380
#define T_Z80_RR 381
#define T_Z80_RRA 382
#define T_Z80_RRC 383
#define T_Z80_RRCA 384
#define T_Z80_SBC 385
#define T_Z80_SCF 386
#define T_Z80_STOP 387
#define T_Z80_SLA 388
#define T_Z80_SRA 389
#define T_Z80_SRL 390
#define T_Z80_SUB 391
#define T_Z80_SWAP 392
#define T_Z80_XOR 393
#define T_MODE_A 394
#define T_MODE_B 395
#define T_MODE_C 396
#define T_MODE_C_IND 397
#define T_MODE_D 398
#define T_MODE_E 399
#define T_MODE_H 400
#define T_MODE_L 401
#define T_MODE_AF 402
#define T_MODE_BC 403
#define T_MODE_BC_IND 404
#define T_MODE_DE 405
#define T_MODE_DE_IND 406
#define T_MODE_SP 407
#define T_MODE_SP_IND 408
#define T_MODE_HL 409
#define T_MODE_HL_IND 410
#define T_MODE_HL_INDDEC 411
#define T_MODE_HL_INDINC 412
#define T_CC_NZ 413
#define T_CC_Z 414
#define T_CC_NC 415
#define YYERRCODE 256
typedef short YYINT;
static const YYINT yylhs[] = {                           -1,
    0,   18,   18,   19,   19,   19,   19,   19,   20,   20,
   20,   20,   26,   22,   25,   25,   25,   27,   24,   24,
   24,   24,   24,   24,   24,   23,   23,   23,   23,   23,
   23,   23,   23,   23,   23,   23,   23,   23,   23,   23,
   23,   23,   23,   23,   23,   23,   23,   23,   23,   23,
   23,   23,   23,   23,   65,   63,   64,   64,   66,   61,
   62,   59,   60,   56,   57,   55,   54,   34,   33,   51,
   50,   32,   31,   30,   48,   45,   46,   47,   71,   58,
   70,   70,   72,   42,   73,   73,   74,   43,   75,   75,
   76,   44,   77,   77,   78,   28,   29,   35,   52,   52,
   53,   53,   37,   38,   36,   39,   40,   41,    3,   67,
   67,   79,   79,   79,   68,   68,   80,   80,   69,   69,
   81,   81,    6,    4,    5,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,   82,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,   83,    2,    2,
    2,    2,    2,    8,    8,    8,    8,    8,   49,   49,
   49,   49,    7,    7,    7,    7,    7,    7,    7,   21,
   21,   21,   21,   21,   21,   21,   21,   21,   21,   21,
   21,   21,   21,   21,   21,   21,   21,   21,   21,   21,
   21,   21,   21,   21,   21,   21,   21,   21,   21,   21,
   21,   21,   21,   21,   21,   21,   21,   21,   21,   21,
   21,   21,   21,   21,   21,   21,   84,   84,   85,   85,
   85,   85,   86,   86,   87,   88,   88,   89,   90,   90,
   91,   92,   93,   93,   94,   95,   96,   96,   97,   98,
   98,   99,   99,   99,   99,  100,  100,  103,  103,  103,
  102,  102,  102,  104,  104,  101,  101,  101,  101,  101,
  101,  101,  101,  136,  136,  136,  137,  137,  132,  132,
  133,  134,  138,  138,  139,  139,  139,  135,  105,  106,
  106,  107,  108,  109,  110,  110,  111,  112,  113,  114,
  115,  116,  117,  118,  119,  120,  121,  121,  122,  123,
  124,  125,  126,  127,  128,  128,  129,  130,  130,   17,
   16,   16,   15,   15,   14,   14,  131,   13,   13,   13,
   13,    9,    9,    9,    9,    9,    9,    9,    9,   12,
   12,   12,   12,   10,   10,   10,   10,   11,   11,   11,
   11,
};
static const YYINT yylen[] = {                            2,
    1,    0,    3,    1,    2,    2,    2,    1,    0,    1,
    2,    3,    0,    3,    0,    1,    3,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    0,    3,    1,    3,    1,    1,
    1,    1,    1,    2,    2,    1,    2,    3,    3,    2,
    1,    3,    3,    3,    2,    2,    2,    2,    0,    3,
    1,    3,    1,    2,    1,    3,    1,    2,    1,    3,
    1,    2,    1,    3,    1,    3,    3,    2,    2,    6,
    4,    4,    2,    2,    2,    2,    1,    1,    1,    1,
    3,    0,    1,    1,    1,    3,    0,    1,    1,    3,
    0,    1,    1,    1,    1,    1,    1,    1,    2,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
    3,    3,    3,    3,    3,    3,    3,    2,    2,    2,
    4,    0,    5,    4,    4,    4,    6,    6,    4,    4,
    4,    4,    4,    4,    6,    6,    6,    4,    3,    1,
    1,    1,    2,    3,    3,    3,    3,    3,    3,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
    3,    3,    2,    2,    2,    4,    4,    4,    6,    6,
    4,    4,    4,    4,    4,    4,    6,    0,    5,    6,
    6,    4,    3,    1,    8,    6,    4,    4,    4,    7,
    9,   12,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    2,    2,    2,    2,
    2,    4,    2,    2,    4,    2,    4,    1,    2,    2,
    1,    1,    2,    2,    1,    1,    4,    4,    1,    2,
    2,    2,    4,    2,    2,    2,    4,    4,    4,    4,
    4,    4,    4,    4,    4,    1,    1,    1,    1,    1,
    1,    1,    1,    7,    5,    4,    4,    4,    4,    4,
    4,    4,    4,    4,    4,    4,    4,    4,    1,    2,
    2,    2,    2,    4,    1,    2,    1,    2,    1,    2,
    1,    2,    1,    2,    1,    2,    2,    2,    1,    4,
    2,    2,    2,    1,    2,    2,    2,    2,    2,    3,
    1,    3,    1,    3,    1,    3,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,
};
static const YYINT yydefred[] = {                         2,
    0,    0,    0,    0,    0,    8,   19,   20,   21,   22,
   23,   24,   25,    0,    0,    0,    0,    0,    0,    0,
    3,   13,    0,    0,    0,    0,    0,    0,  107,  108,
    0,    0,    0,    0,    0,    0,    0,    0,   71,    0,
    0,    0,    0,   66,    0,    0,   79,   62,   63,   60,
   61,   55,    0,    0,    0,    0,    0,  288,    0,  291,
  292,    0,  295,  296,    0,  299,    0,    0,    0,    0,
    0,    0,    0,  339,    0,    0,    0,    0,    0,  347,
    0,    0,  349,    0,  351,    0,  353,    0,  355,    0,
  359,  364,    0,    0,    0,    0,    0,    0,    5,    6,
    7,   26,   27,   28,   29,   30,   31,   32,   33,   34,
   35,   36,   37,   38,   39,   40,   41,   42,   43,   44,
   45,   46,   47,   48,   49,   50,   51,   52,   53,   54,
  230,  231,  232,  233,  234,  235,  236,  237,  238,  239,
  240,  241,  242,  243,  244,  245,  246,  247,  248,  249,
  250,  251,  252,  253,  254,  255,  256,  257,  258,  259,
  260,  261,  262,  263,  264,  265,  266,  267,  268,  269,
  270,  271,  272,  273,  274,  275,  276,  316,  317,  318,
  319,  320,  321,  322,  323,  171,  214,    0,    0,    0,
    0,  208,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  172,    0,   69,    0,    0,    0,
   68,   12,    0,    0,    0,   98,    0,  103,    0,    0,
   87,   84,    0,   91,   88,    0,   95,   92,    0,  127,
    0,    0,    0,    0,  152,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  126,    0,    0,  113,    0,   76,    0,    0,    0,
  118,  128,   77,    0,    0,   78,    0,    0,    0,    0,
    0,    0,   64,   65,    0,    0,    0,  382,  383,  384,
  385,  386,  387,  388,  375,  373,  277,  278,  394,  395,
    0,    0,  371,  279,  280,  281,  283,  284,    0,  381,
  378,  379,  380,  286,    0,  289,  290,  389,  397,  396,
  293,  294,    0,    0,  300,  301,  305,  304,  302,    0,
    0,  306,    0,    0,  398,  399,    0,    0,  401,  400,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  340,  341,  393,  390,  391,  392,  342,  343,    0,
  346,  356,  348,  350,  352,  354,  357,  358,  361,  362,
  363,  365,  366,  367,  368,  369,  173,  193,  194,  195,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   18,   14,    0,  377,    0,    0,    0,    0,  129,  148,
  149,  150,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   83,   80,    0,   59,   56,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  184,  213,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  190,  191,  192,    0,  360,   86,   90,   94,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  169,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  145,  146,  147,  111,  116,  120,  223,  224,
  225,  226,  227,  228,  229,    0,    0,    0,    0,    0,
    0,  376,  374,  282,  372,  285,  287,  298,  297,  303,
  307,  331,  327,  328,    0,    0,  326,  370,  335,  333,
  334,  336,  337,  338,  332,  330,  329,  309,  310,  308,
  312,  313,  311,  314,  315,  344,    0,  201,  202,  203,
  204,  205,  206,    0,    0,    0,  196,  197,  198,    0,
    0,    0,  212,    0,  217,  218,   17,    0,  151,  159,
  160,  161,  162,  163,  164,    0,    0,    0,  154,  155,
  156,    0,    0,  168,    0,    0,    0,   82,   58,  325,
    0,  209,    0,    0,    0,    0,    0,    0,    0,  153,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  207,
  199,  200,  210,  211,    0,  216,  165,  157,  158,  166,
  167,    0,    0,  324,    0,    0,    0,  215,  221,    0,
    0,    0,  222,
};
static const YYINT yydgoto[] = {                          1,
  264,  224,  225,  295,  271,  332,  596,  215,  296,  303,
  344,  358,  315,  297,  298,  306,  345,    2,    4,    5,
   99,  100,  101,    6,  422,  223,  423,    7,    8,    9,
   10,   11,   12,   13,  102,  103,  104,  105,  106,  107,
  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,
  118,  119,  120,  121,  122,  123,  124,  125,  126,  127,
  128,  129,  130,  479,  286,  480,  267,  273,  276,  476,
  285,  477,  232,  233,  235,  236,  238,  239,  268,  274,
  277,  433,  381,  131,  132,  133,  134,  135,  136,  137,
  138,  139,  140,  141,  142,  143,  144,  145,  146,  147,
  148,  149,  150,  151,  152,  153,  154,  155,  156,  157,
  158,  159,  160,  161,  162,  163,  164,  165,  166,  167,
  168,  169,  170,  171,  172,  173,  174,  175,  176,  177,
  425,  178,  179,  180,  181,  182,  183,  184,  185,
};
static const YYINT yysindex[] = {                         0,
    0, -291,  -24,   15, 2933,    0,    0,    0,    0,    0,
    0,    0,    0, 1519, 1519, -221, 1519, 1519, 1519,  -46,
    0,    0, 1519, -221, 1519, -221, 1519, 1519,    0,    0,
 -244, -240, -238, 1338, 1519, 1338, 1338, -221,    0, 1519,
 -221, 1519, -221,    0, -221, -221,    0,    0,    0,    0,
    0,    0,  991,  930,  991, 1519,  787,    0,  991,    0,
    0, -246,    0,    0, -308,    0, -246,  296,  787,  205,
 -363, -313,  -89,    0,  991, -370, -370, 1519, -270,    0,
 1338, -193,    0, -193,    0, -193,    0, -193,    0,  991,
    0,    0, -193, -193, -193,  991, -193,  991,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1519, 1519, 1519,
 1519,    0,   47,   66,   68,   83,   85,   93,   99,  101,
  111,  116,  118,  120,  138,  158,  164,  170,  172,  176,
  192, -185, 1519,  -80,    0,  -80,    0,  -80,  -80,  -80,
    0,    0, -164,  -80,  186,    0,  -80,    0,  -80,  -80,
    0,    0,  189,    0,    0,  209,    0,    0,  210,    0,
 1338, 1338, 1338, 1338,    0,  242,  249,  254,  255,  283,
  292,  293,  294,  295,  297,  298,  299,  300,  302,  303,
  304,    0, 1338, 3141,    0,    0,    0,  301,  -80, 3141,
    0,    0,    0,  320, 3141,    0,  321,  322,  -80,  323,
  -80,  336,    0,    0,  105,  157,  186,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  186,  186,    0,    0,    0,    0,    0,    0,  186,    0,
    0,    0,    0,    0,  186,    0,    0,    0,    0,    0,
    0,    0,  186,  186,    0,    0,    0,    0,    0,  186,
 3141,    0,  186,  186,    0,    0,  186,  186,    0,    0,
 1338,  186,  186,  186,  186,  186,  186,  186,  186,  186,
  186,    0,    0,    0,    0,    0,    0,    0,    0,  186,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  376, 1519, 1519, 1519, 1519, 1519, 1519, 1519, 1519, 1519,
 1519, 1519, 1519, -221, -221, -221, -221, -221, -221, -221,
  115,  -25, 1519, 1519, 1519, 1519, 1519, 1519, 1519, 1519,
 1519, 1519, 1519, 1519, 1519, 1519, 1519, 1519, 1519, 1519,
    0,    0,  374,    0, -193, -244, -240, -238,    0,    0,
    0,    0,  380,  133, 1519, 1519, 1519, 1519, 1519, 1519,
 1519, 1519, 1519, 1519, 1519, 1519, -221, -221, -221, 1900,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338,  177, 1519, 1519,    0,    0,  394,    0,    0,  398,
 1125, 1338, -194, -193, 1338,   36,   50, 1338, 1338,   52,
 1190, 1259,  356,  841, 1338,   92, -315, -281,   95, -181,
   96,  396,   97, -193,  191, 1929, 2049, 2070, 2088, 2106,
 2137, 2723, 2793, 2814, 2166, 2194, 2222,  448,  450,  453,
  457,  456,  469,  470,    0,    0,  406,  406,  406,  -50,
  -50,  -50,  -50,  -50, -104, -104,   11,   11,   11, -139,
 -139,    0,    0,    0, -164,    0,    0,    0,    0,  213,
  471, 2252, 2286, 2343, 2374, 2403, 2431, 2845, 2874, 2902,
 2459, 2489, 2523,  472,  481,  487,    0,  886,  886,  886,
   38,   38,   38,   38,   38,   54,   54,   27,   27,   27,
   41,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  -37, 2930,  -80,    0,  105,
  157,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1338,  143,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  510,    0,    0,    0,
    0,    0,    0, 1519, 1519, 1519,    0,    0,    0, -221,
 -221, 1519,    0, -221,    0,    0,    0,  511,    0,    0,
    0,    0,    0,    0,    0, 1519, 1519, 1519,    0,    0,
    0, -221, -221,    0,  276, 1519, 1519,    0,    0,    0,
 1338,    0, 2541, 2559, 2580,  516,  517, 2957,  518,    0,
 2611, 2640, 2668,  519,  520,  477, 1169,  -80,  478,    0,
    0,    0,    0,    0, 1519,    0,    0,    0,    0,    0,
    0, 1519,  528,    0, 2696, 2745,  282,    0,    0,  505,
 1519, 2766,    0,
};
static const YYINT yyrindex[] = {                         0,
    0,  433,  629,    0,  587,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  537,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   12,    0,   17,   19,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  597,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   86,    0,  603,    0,  608,    0,  609,  610,  611,
    0,    0,  612,  579,    0,    0,  614,    0,  615,  616,
    0,    0,  617,    0,    0,  618,    0,    0,  623,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   -2,    0,   -1,    0,  624,  625,    9,
    0,    0,    0,  626,   26,    0,  627,    0,  630,  631,
  632,    0,    0,    0,    0,    0,  633,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  634,  635,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  636,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  638,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   12,   17,
   19,    0,    0,    0,    0,    0,  639,    0,    0,  640,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   31, 1138, 1251,  121,
  190, 1842, 1892, 1959, 1726, 1755, 1564, 1600, 1628, 1049,
 1103,    0,    0,    0,  612,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1521, 2040, 2045,
 1967, 1987, 1995, 2017, 2025, 1810, 1832, 1650, 1682, 1696,
 1416, 1488,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  641,    0,  645,  454,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  646,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  647,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,
};
static const YYINT yygindex[] = {                         0,
   78,  -14,  -32,   32,  -51,  169,    0,   14,   23,  -47,
  165,  583,   53,  -10,   44,    0,  -55,    0,    0,    0,
    0,    0,    0,    0,  122,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   60,    0,    0,  197,  214,  217,   89,
    0,    0,  259,    0,  263,    0,  265,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 3040,    0,    0,    0,    0,    0,    0,    0,    0,
};
#define YYTABLESIZE 3418
static const YYINT yytable[] = {                        214,
  216,  341,  218,  219,  220,  314,  675,  124,  114,    3,
  227,  222,  229,  230,  322,  526,  329,  351,  125,  326,
  269,  112,  343,  309,   21,  279,  117,  281,  121,  217,
  346,  354,  355,   20,  356,  122,  187,  226,  357,  228,
  174,  124,  114,  304,  307,  360,  347,  266,  316,  272,
  272,  278,  125,  676,  280,  112,  282,  231,  283,  284,
  117,  234,  121,  237,  352,  265,  272,  272,  272,  122,
  272,  174,  272,  207,  174,  209,  210,  211,  626,  367,
  348,  272,  272,  401,  321,  372,  382,  375,  272,  325,
  124,  627,  342,  421,  272,  170,  349,  305,  308,  323,
  324,  125,  317,  272,  363,  383,  364,  384,  365,  272,
  366,  272,  362,  270,  275,  369,  370,  371,  353,  374,
  330,  333,  385,  174,  386,  310,  170,  628,  629,  170,
  177,  361,  387,  368,  270,  418,  419,  420,  388,  373,
  389,  376,  311,  312,  313,  270,  331,  318,  288,  289,
  390,  290,  291,  292,  293,  391,  299,  392,  300,  393,
  319,  177,  320,  294,  177,  413,  414,  415,  416,  417,
  418,  419,  420,  377,  378,  379,  380,  394,  170,  403,
  404,  405,  406,  407,  408,  409,  410,  411,  412,  413,
  414,  415,  416,  417,  418,  419,  420,  395,  402,  178,
  318,  288,  289,  396,  290,  291,  292,  293,  299,  397,
  300,  398,  319,  177,  320,  399,  294,  411,  412,  413,
  414,  415,  416,  417,  418,  419,  420,  631,  632,  424,
  178,  400,  426,  178,  403,  404,  405,  406,  407,  408,
  409,  410,  411,  412,  413,  414,  415,  416,  417,  418,
  419,  420,  427,  428,  272,  272,  272,  272,  128,  128,
  128,  128,  128,  128,  128,  128,  128,  128,  128,  128,
  128,  128,  128,  128,  128,  128,  272,  221,   14,   15,
   16,  434,  178,  416,  417,  418,  419,  420,  435,  493,
  174,  174,  174,  436,  437,  341,   17,   18,   19,  464,
  465,  466,  467,  468,  350,  459,  460,  461,  462,  463,
  464,  465,  466,  467,  468,  466,  467,  468,  429,  430,
  431,  432,  438,  461,  462,  463,  464,  465,  466,  467,
  468,  439,  440,  441,  442,  263,  443,  444,  445,  446,
  450,  447,  448,  449,  469,  170,  170,  170,  170,  170,
  170,  170,  170,  170,  272,  170,  170,  170,  170,  170,
  170,  170,  170,  470,  471,  472,  473,  506,  507,  508,
  509,  510,  511,  512,  513,  514,  515,  516,  517,  474,
  177,  177,  177,  177,  177,  177,  177,  177,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,  544,  475,  518,  519,  520,
  521,  522,  523,  524,  478,  505,  525,  545,  270,  550,
  552,  553,  554,  555,  556,  557,  558,  559,  560,  561,
  562,  563,    1,  607,  551,  605,  610,  600,  623,  614,
  617,  601,    9,  624,  608,  612,  634,  546,  618,  178,
  178,  178,  178,  178,  178,  178,  178,  609,  597,  598,
  564,  565,  566,  101,  272,  272,  272,  272,  272,  272,
  272,  272,  272,  272,  272,  272,  272,  272,  272,  272,
  272,  272,  266,  272,  272,  625,  341,  599,  630,  633,
  635,  650,  637,  651,  272,  272,  652,  653,  272,  654,
  265,  272,  272,  603,  272,  272,  606,  272,  272,  655,
  656,  659,  602,  604,  658,  672,  621,  589,  590,  591,
  592,  593,  594,  595,  673,  620,  636,  674,  568,  569,
  570,  571,  572,  573,  574,  575,  576,  577,  578,  579,
  580,  581,  582,  583,  584,  585,   11,  270,  275,  681,
  682,  690,  240,  187,  241,  696,  703,  704,  706,  710,
  711,  720,  270,  242,  243,  270,  331,  712,  270,  270,
  714,  717,  270,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  207,  261,  209,  210,  211,  721,    4,  262,  318,  288,
  289,  334,  290,  291,  292,  293,  345,  299,  335,  300,
  336,  337,   96,  338,  294,  339,  340,   97,   74,   73,
   72,   15,  109,  105,  104,  106,   85,   89,  272,  683,
  684,  685,   93,  110,   75,  115,  119,  688,   10,   70,
   99,   67,  389,  397,  396,  123,  680,   16,   81,   57,
  219,  691,  692,  693,  102,  100,  220,  611,  622,  359,
  679,  697,  698,  686,  687,  586,  657,  689,  406,  407,
  408,  409,  410,  411,  412,  413,  414,  415,  416,  417,
  418,  419,  420,  587,  547,  694,  695,  588,  678,  548,
  715,  310,  549,    0,  272,    0,    0,  716,    0,    0,
    0,    0,    0,    0,  327,  328,  722,    0,  311,  312,
  313,    0,  699,  172,  172,  172,  172,  172,  172,  172,
  172,  172,  172,  172,  172,  172,  172,  172,  172,  172,
  172,    0,    0,    0,    9,    0,    9,    0,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    0,    0,    0,    0,    0,    9,    9,
    9,    9,    9,    9,    0,    9,    9,    9,    9,    9,
    9,    9,    9,    0,    0,    0,    0,    0,    0,    0,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,  263,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   11,    0,
   11,    0,   11,   11,   11,   11,   11,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,    0,    0,    0,
    0,    0,   11,   11,   11,   11,   11,   11,    0,   11,
   11,   11,   11,   11,   11,   11,   11,    0,    0,    0,
  263,    0,    0,    0,   11,   11,   11,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
   10,  341,    0,    0,   10,   10,   10,   10,   10,   10,
   10,   10,   10,   10,   10,   10,   10,   10,   10,    0,
    0,    0,    0,    0,   10,   10,   10,   10,   10,   10,
    0,   10,   10,   10,   10,   10,   10,   10,   10,  263,
    0,    0,    0,    0,    0,    0,   10,   10,   10,   10,
   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,
   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,
   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,
   10,   10,   10,   10,   10,   10,   10,   10,   10,   10,
   10,   10,    0,    0,    0,    0,    0,    0,    0,    0,
  263,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  240,  187,  241,    0,    0,    0,    0,
    0,    0,    0,    0,  242,  243,    0,    0,  188,    0,
    0,    0,    0,    0,  244,  245,  246,  247,  248,  249,
  250,  251,  252,  253,  254,  255,  256,  257,  258,  259,
  260,  207,  261,  209,  210,  211,    0,    0,  262,  188,
    0,    0,  188,    0,    0,    0,    0,  240,  187,  241,
    0,    0,    0,    0,    0,    0,    0,    0,  242,  243,
    0,    0,  189,    0,    0,    0,    0,    0,  244,  245,
  246,  247,  248,  249,  250,  251,  252,  253,  254,  255,
  256,  257,  258,  259,  260,  207,  261,  209,  210,  211,
    0,  188,  262,  189,    0,    0,  189,  175,  454,  455,
  456,  457,  458,  459,  460,  461,  462,  463,  464,  465,
  466,  467,  468,    0,  263,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  175,    0,
    0,  175,  310,    0,    0,    0,  240,  187,  241,    0,
    0,    0,    0,    0,    0,  189,    0,  242,  243,  311,
  312,  313,    0,    0,    0,    0,    0,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  207,  261,  209,  210,  211,  263,
  175,  262,    0,    0,  318,  288,  289,  619,  290,  291,
  292,  293,    0,    0,  335,    0,  336,  240,  187,  241,
  294,  339,  340,    0,    0,    0,    0,    0,  242,  243,
  176,  713,    0,    0,    0,    0,    0,    0,  244,  245,
  246,  247,  248,  249,  250,  251,  252,  253,  254,  255,
  256,  257,  258,  259,  260,  207,  261,  209,  210,  211,
    0,  176,  262,    0,  176,    0,    0,    0,  263,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  188,  188,
  188,  188,  188,  188,  188,  188,  188,  188,  188,  188,
  188,  188,  188,  287,  288,  289,    0,  290,  291,  292,
  293,    0,  299,    0,  300,    0,  301,    0,  302,  294,
    0,    0,    0,  176,    0,    0,    0,    0,    0,  616,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  189,  189,  189,  189,  189,  189,  189,  189,
  189,  189,  189,  189,  189,  189,  189,  263,    0,    0,
    0,  240,  187,  241,  287,  288,  289,    0,  290,  291,
  292,  293,  242,  243,    0,    0,    0,  175,  175,  175,
  294,    0,  244,  245,  246,  247,  248,  249,  250,  251,
  252,  253,  254,  255,  256,  257,  258,  259,  260,  207,
  261,  209,  210,  211,    0,  143,  262,    0,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  240,  187,  241,    0,
    0,    0,    0,    0,    0,    0,  143,  242,  243,  143,
    0,    0,    0,    0,    0,    0,    0,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  207,  261,  209,  210,  211,    0,
    0,  262,    0,    0,    0,    0,    0,  144,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  143,    0,
  176,  176,  176,    0,    0,  240,  187,  241,  318,  288,
  289,    0,  290,  291,  292,  293,  242,  243,  144,    0,
  130,  144,    0,    0,  294,    0,  244,  245,  246,  247,
  248,  249,  250,  251,  252,  253,  254,  255,  256,  257,
  258,  259,  260,  207,  261,  209,  210,  211,  213,    0,
  262,  130,    0,    0,  130,    0,    0,    0,    0,    0,
    0,    0,    0,  186,    0,    0,    0,    0,    0,    0,
  144,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  240,  187,  241,    0,  613,    0,
    0,    0,    0,    0,  186,  242,  243,  186,    0,  185,
    0,    0,    0,  130,    0,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  207,  261,  209,  210,  211,  187,    0,  262,
  185,    0,    0,  185,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  186,    0,    0,  141,
    0,    0,    0,    0,    0,  615,    0,    0,  187,    0,
    0,  187,    0,    0,    0,  143,  143,  143,  143,  143,
  143,  143,  143,  143,  143,  143,  143,  143,  143,  143,
  141,  140,  185,  141,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  142,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  187,    0,  140,    0,    0,  140,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  182,  142,    0,    0,  142,
    0,    0,  141,    0,    0,    0,    0,  144,  144,  144,
  144,  144,  144,  144,  144,  144,  144,  144,  144,  144,
  144,  144,    0,    0,  183,    0,  182,    0,    0,  182,
    0,    0,    0,    0,  140,  186,  187,  188,    0,    0,
  130,  130,  130,    0,    0,    0,  189,  190,  142,    0,
    0,    0,    0,    0,    0,  183,  191,  192,  183,  193,
  194,  195,  196,  197,  198,  199,  200,  201,  202,  203,
  204,  205,  206,  207,  208,  209,  210,  211,  182,  138,
  212,    0,    0,  186,  186,  186,  186,  186,  186,  186,
  186,  186,  186,  186,  186,  186,    0,    0,    0,    0,
    0,  139,    0,    0,    0,    0,    0,  183,    0,    0,
  138,  179,    0,  138,    0,    0,    0,    0,    0,  185,
  185,  185,  185,  185,  185,  185,  185,  185,  185,  185,
  185,  185,  139,    0,    0,  139,    0,    0,    0,    0,
    0,    0,  179,    0,    0,  179,    0,  187,  187,  187,
  187,  187,  187,  187,  187,  187,  187,  187,  187,  187,
    0,  180,  138,    0,    0,    0,    0,    0,    0,  141,
  141,  141,  141,  141,  141,  141,  141,  141,  141,  141,
  141,  141,    0,    0,  139,    0,    0,    0,    0,    0,
    0,    0,  180,    0,  179,  180,    0,    0,    0,    0,
  567,  140,  140,  140,  140,  140,  140,  140,  140,  140,
  140,  140,  140,  140,    0,  142,  142,  142,  142,  142,
  142,  142,  142,  142,  142,  142,  142,  142,  181,  638,
    0,    0,    0,    0,    0,    0,  133,    0,    0,    0,
    0,    0,    0,    0,  180,  182,  182,  182,  182,  182,
  182,  182,  182,  182,  182,    0,  134,    0,    0,  181,
    0,    0,  181,    0,  135,    0,    0,  133,    0,    0,
  133,    0,    0,    0,  183,  183,  183,  183,  183,  183,
  183,  183,  183,  183,    0,    0,  136,  134,    0,    0,
  134,    0,    0,    0,  137,  135,    0,    0,  135,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  131,
    0,  181,    0,    0,  132,    0,    0,  136,    0,  133,
  136,    0,    0,    0,    0,  137,    0,    0,  137,  138,
  138,  138,  138,  138,  138,  138,  138,  138,  138,  134,
  131,    0,    0,  131,    0,  132,    0,  135,  132,  639,
    0,  139,  139,  139,  139,  139,  139,  139,  139,  139,
  139,  179,  179,  179,  179,  179,  179,  179,  179,  136,
  640,    0,    0,    0,    0,    0,    0,  137,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  641,    0,
    0,    0,  131,    0,    0,    0,    0,  132,    0,    0,
    0,    0,    0,    0,    0,    0,  642,    0,    0,    0,
    0,  180,  180,  180,  180,  180,  180,  180,  180,  451,
  452,  453,  454,  455,  456,  457,  458,  459,  460,  461,
  462,  463,  464,  465,  466,  467,  468,  643,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  647,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  181,  181,
  181,  181,  181,  181,  181,  181,  133,  133,  133,  133,
  133,  133,  133,  133,  648,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  134,  134,  134,  134,
  134,  134,  134,  134,  135,  135,  135,  135,  135,  135,
  135,  135,  649,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  136,  136,  136,  136,
  136,  136,  136,  136,  137,  137,  137,  137,  137,  137,
  137,  137,  660,    0,    0,    0,    0,    0,    0,  131,
  131,  131,    0,    0,  132,  132,  132,    0,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  661,    0,    0,  403,
  404,  405,  406,  407,  408,  409,  410,  411,  412,  413,
  414,  415,  416,  417,  418,  419,  420,  403,  404,  405,
  406,  407,  408,  409,  410,  411,  412,  413,  414,  415,
  416,  417,  418,  419,  420,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  412,  413,  414,  415,  416,  417,
  418,  419,  420,  662,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  403,  404,  405,  406,
  407,  408,  409,  410,  411,  412,  413,  414,  415,  416,
  417,  418,  419,  420,  663,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  412,  413,  414,  415,  416,  417,
  418,  419,  420,  664,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,  415,  416,  417,  418,  419,
  420,  665,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  412,  413,  414,  415,  416,  417,  418,  419,  420,  669,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  412,  413,  414,  415,  416,  417,  418,  419,  420,  670,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  412,  413,  414,  415,  416,  417,
  418,  419,  420,  671,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  700,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  701,
    0,    0,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  414,  415,  416,  417,  418,  419,  420,
  702,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,  415,  416,  417,  418,  419,
  420,  707,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  414,  415,  416,  417,  418,  419,  420,
  708,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  403,  404,  405,  406,  407,  408,  409,  410,  411,  412,
  413,  414,  415,  416,  417,  418,  419,  420,  709,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  718,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  644,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  414,  415,  416,  417,  418,  419,  420,
  403,  404,  405,  406,  407,  408,  409,  410,  411,  412,
  413,  414,  415,  416,  417,  418,  419,  420,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  645,  719,    0,  403,
  404,  405,  406,  407,  408,  409,  410,  411,  412,  413,
  414,  415,  416,  417,  418,  419,  420,  646,  723,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  403,  404,  405,  406,  407,  408,  409,  410,  411,  412,
  413,  414,  415,  416,  417,  418,  419,  420,  666,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  403,
  404,  405,  406,  407,  408,  409,  410,  411,  412,  413,
  414,  415,  416,  417,  418,  419,  420,  667,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  403,  404,  405,
  406,  407,  408,  409,  410,  411,  412,  413,  414,  415,
  416,  417,  418,  419,  420,  668,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  412,  413,  414,  415,  416,  417,
  418,  419,  420,  677,    0,    0,    0,    0,    0,    0,
    0,    0,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  414,  415,  416,  417,  418,  419,  420,
  705,    0,    0,    0,  403,  404,  405,  406,  407,  408,
  409,  410,  411,  412,  413,  414,  415,  416,  417,  418,
  419,  420,    0,    0,    0,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  412,  413,  414,  415,  416,  417,
  418,  419,  420,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  414,  415,  416,  417,  418,  419,  420,
    0,    0,    0,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,  415,  416,  417,  418,  419,
  420,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  403,  404,  405,  406,  407,  408,
  409,  410,  411,  412,  413,  414,  415,  416,  417,  418,
  419,  420,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,  415,  416,  417,  418,  419,
  420,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  412,  413,  414,  415,  416,  417,  418,  419,  420,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  403,
  404,  405,  406,  407,  408,  409,  410,  411,  412,  413,
  414,  415,  416,  417,  418,  419,  420,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  403,  404,  405,  406,
  407,  408,  409,  410,  411,  412,  413,  414,  415,  416,
  417,  418,  419,  420,   22,    0,   23,    0,   24,   25,
   26,   27,   28,   29,   30,   31,   32,   33,   34,   35,
   36,   37,   38,    0,    0,    0,    0,    0,   39,   40,
   41,   42,   43,   44,    0,   45,   46,   47,   48,   49,
   50,   51,   52,    0,    0,    0,    0,    0,    0,    0,
   53,   54,   55,   56,   57,   58,   59,   60,   61,   62,
   63,   64,   65,   66,   67,   68,   69,   70,   71,   72,
   73,   74,   75,   76,   77,   78,   79,   80,   81,   82,
   83,   84,   85,   86,   87,   88,   89,   90,   91,   92,
   93,   94,   95,   96,   97,   98,  481,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  482,  483,    0,    0,    0,    0,    0,    0,  484,    0,
    0,    0,    0,    0,  485,    0,    0,    0,    0,    0,
    0,    0,  486,  487,    0,    0,    0,    0,    0,  488,
    0,    0,  489,  490,    0,    0,  491,  492,    0,    0,
    0,  494,  495,  496,  497,  498,  499,  500,  501,  502,
  503,    0,    0,    0,    0,    0,    0,    0,    0,  504,
  451,  452,  453,  454,  455,  456,  457,  458,  459,  460,
  461,  462,  463,  464,  465,  466,  467,  468,
};
static const YYINT yycheck[] = {                         14,
   15,   91,   17,   18,   19,   57,   44,   10,   10,  301,
   25,   58,   27,   28,   62,   41,   68,   73,   10,   67,
   35,   10,   70,   56,   10,   40,   10,   42,   10,   16,
  394,  402,  403,   58,  405,   10,  258,   24,  409,   26,
   10,   44,   44,   54,   55,   78,  410,   34,   59,   36,
   37,   38,   44,   91,   41,   44,   43,  302,   45,   46,
   44,  302,   44,  302,   75,   34,   53,   54,   55,   44,
   57,   41,   59,  295,   44,  297,  298,  299,  394,   90,
  394,   68,   69,  269,   62,   96,   40,   98,   75,   67,
   93,  407,   70,  258,   81,   10,  410,   54,   55,  408,
  409,   93,   59,   90,   82,   40,   84,   40,   86,   96,
   88,   98,   81,   36,   37,   93,   94,   95,   75,   97,
   68,   69,   40,   93,   40,  396,   41,  409,  410,   44,
   10,   79,   40,   90,   57,  275,  276,  277,   40,   96,
   40,   98,  413,  414,  415,   68,   69,  394,  395,  396,
   40,  398,  399,  400,  401,   40,  403,   40,  405,   40,
  407,   41,  409,  410,   44,  270,  271,  272,  273,  274,
  275,  276,  277,  188,  189,  190,  191,   40,   93,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,   40,  213,   10,
  394,  395,  396,   40,  398,  399,  400,  401,  403,   40,
  405,   40,  407,   93,  409,   40,  410,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,  409,  410,   44,
   41,   40,   44,   44,  260,  261,  262,  263,  264,  265,
  266,  267,  268,  269,  270,  271,  272,  273,  274,  275,
  276,  277,   44,   44,  241,  242,  243,  244,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,  263,  324,  303,  304,
  305,   40,   93,  273,  274,  275,  276,  277,   40,  341,
  260,  261,  262,   40,   40,   91,  321,  322,  323,  273,
  274,  275,  276,  277,  394,  268,  269,  270,  271,  272,
  273,  274,  275,  276,  277,  275,  276,  277,  241,  242,
  243,  244,   40,  270,  271,  272,  273,  274,  275,  276,
  277,   40,   40,   40,   40,   40,   40,   40,   40,   40,
  263,   40,   40,   40,   44,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  341,  270,  271,  272,  273,  274,
  275,  276,  277,   44,   44,   44,   44,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,  392,  393,   44,
  260,  261,  262,  263,  264,  265,  266,  267,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,  414,
  415,  416,  417,  418,  419,  420,  302,  394,  395,  396,
  397,  398,  399,  400,  258,   40,  302,   44,  341,   40,
  435,  436,  437,  438,  439,  440,  441,  442,  443,  444,
  445,  446,    0,  485,  302,  483,  488,   44,  494,  491,
  492,   44,   10,  495,  409,  394,  502,  425,   93,  260,
  261,  262,  263,  264,  265,  266,  267,  408,  473,  474,
  447,  448,  449,   10,  451,  452,  453,  454,  455,  456,
  457,  458,  459,  460,  461,  462,  463,  464,  465,  466,
  467,  468,  469,  470,  471,  394,   91,  474,  394,  394,
  394,   44,  302,   44,  481,  482,   44,   41,  485,   44,
  469,  488,  489,  481,  491,  492,  484,  494,  495,   41,
   41,   41,  481,  482,  302,   44,  494,  341,  342,  343,
  344,  345,  346,  347,   44,  494,  504,   41,  451,  452,
  453,  454,  455,  456,  457,  458,  459,  460,  461,  462,
  463,  464,  465,  466,  467,  468,   10,  470,  471,  407,
   41,   41,  257,  258,  259,  280,   41,   41,   41,   41,
   41,  280,  485,  268,  269,  488,  489,   91,  491,  492,
   93,   44,  495,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,  292,  293,  294,
  295,  296,  297,  298,  299,   91,   10,  302,  394,  395,
  396,  397,  398,  399,  400,  401,   10,  403,  404,  405,
  406,  407,   10,  409,  410,  411,  412,   10,   10,   10,
   10,   10,   44,   10,   10,   10,   10,   10,  615,  644,
  645,  646,   10,   10,   10,   10,   10,  652,   10,   10,
   10,   10,   10,   10,   10,   10,  615,   10,   10,   10,
   10,  666,  667,  668,   10,   10,   10,  489,  494,   77,
  601,  676,  677,  650,  651,  469,  545,  654,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,  470,  426,  672,  673,  471,  600,  427,
  705,  396,  428,   -1,  681,   -1,   -1,  712,   -1,   -1,
   -1,   -1,   -1,   -1,  409,  410,  721,   -1,  413,  414,
  415,   -1,  681,  260,  261,  262,  263,  264,  265,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,   -1,   -1,   -1,  302,   -1,  304,   -1,  306,  307,
  308,  309,  310,  311,  312,  313,  314,  315,  316,  317,
  318,  319,  320,   -1,   -1,   -1,   -1,   -1,  326,  327,
  328,  329,  330,  331,   -1,  333,  334,  335,  336,  337,
  338,  339,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,  354,  355,  356,  357,
  358,  359,  360,  361,  362,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,  375,  376,  377,
  378,  379,  380,  381,  382,  383,  384,  385,  386,  387,
  388,  389,  390,  391,  392,  393,   40,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  302,   -1,
  304,   -1,  306,  307,  308,  309,  310,  311,  312,  313,
  314,  315,  316,  317,  318,  319,  320,   -1,   -1,   -1,
   -1,   -1,  326,  327,  328,  329,  330,  331,   -1,  333,
  334,  335,  336,  337,  338,  339,  340,   -1,   -1,   -1,
   40,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
  354,  355,  356,  357,  358,  359,  360,  361,  362,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,  375,  376,  377,  378,  379,  380,  381,  382,  383,
  384,  385,  386,  387,  388,  389,  390,  391,  392,  393,
  302,   91,   -1,   -1,  306,  307,  308,  309,  310,  311,
  312,  313,  314,  315,  316,  317,  318,  319,  320,   -1,
   -1,   -1,   -1,   -1,  326,  327,  328,  329,  330,  331,
   -1,  333,  334,  335,  336,  337,  338,  339,  340,   40,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,  354,  355,  356,  357,  358,  359,  360,  361,
  362,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,  375,  376,  377,  378,  379,  380,  381,
  382,  383,  384,  385,  386,  387,  388,  389,  390,  391,
  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   40,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,  258,  259,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  268,  269,   -1,   -1,   10,   -1,
   -1,   -1,   -1,   -1,  278,  279,  280,  281,  282,  283,
  284,  285,  286,  287,  288,  289,  290,  291,  292,  293,
  294,  295,  296,  297,  298,  299,   -1,   -1,  302,   41,
   -1,   -1,   44,   -1,   -1,   -1,   -1,  257,  258,  259,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  268,  269,
   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,  278,  279,
  280,  281,  282,  283,  284,  285,  286,  287,  288,  289,
  290,  291,  292,  293,  294,  295,  296,  297,  298,  299,
   -1,   93,  302,   41,   -1,   -1,   44,   10,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,   -1,   40,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   41,   -1,
   -1,   44,  396,   -1,   -1,   -1,  257,  258,  259,   -1,
   -1,   -1,   -1,   -1,   -1,   93,   -1,  268,  269,  413,
  414,  415,   -1,   -1,   -1,   -1,   -1,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,  292,  293,  294,  295,  296,  297,  298,  299,   40,
   93,  302,   -1,   -1,  394,  395,  396,  397,  398,  399,
  400,  401,   -1,   -1,  404,   -1,  406,  257,  258,  259,
  410,  411,  412,   -1,   -1,   -1,   -1,   -1,  268,  269,
   10,   93,   -1,   -1,   -1,   -1,   -1,   -1,  278,  279,
  280,  281,  282,  283,  284,  285,  286,  287,  288,  289,
  290,  291,  292,  293,  294,  295,  296,  297,  298,  299,
   -1,   41,  302,   -1,   44,   -1,   -1,   -1,   40,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  394,  395,  396,   -1,  398,  399,  400,
  401,   -1,  403,   -1,  405,   -1,  407,   -1,  409,  410,
   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  260,  261,  262,  263,  264,  265,  266,  267,
  268,  269,  270,  271,  272,  273,  274,   40,   -1,   -1,
   -1,  257,  258,  259,  394,  395,  396,   -1,  398,  399,
  400,  401,  268,  269,   -1,   -1,   -1,  260,  261,  262,
  410,   -1,  278,  279,  280,  281,  282,  283,  284,  285,
  286,  287,  288,  289,  290,  291,  292,  293,  294,  295,
  296,  297,  298,  299,   -1,   10,  302,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,  257,  258,  259,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   41,  268,  269,   44,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,  292,  293,  294,  295,  296,  297,  298,  299,   -1,
   -1,  302,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,
  260,  261,  262,   -1,   -1,  257,  258,  259,  394,  395,
  396,   -1,  398,  399,  400,  401,  268,  269,   41,   -1,
   10,   44,   -1,   -1,  410,   -1,  278,  279,  280,  281,
  282,  283,  284,  285,  286,  287,  288,  289,  290,  291,
  292,  293,  294,  295,  296,  297,  298,  299,   40,   -1,
  302,   41,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,  258,  259,   -1,  409,   -1,
   -1,   -1,   -1,   -1,   41,  268,  269,   44,   -1,   10,
   -1,   -1,   -1,   93,   -1,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,  292,
  293,  294,  295,  296,  297,  298,  299,   10,   -1,  302,
   41,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   10,
   -1,   -1,   -1,   -1,   -1,  407,   -1,   -1,   41,   -1,
   -1,   44,   -1,   -1,   -1,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
   41,   10,   93,   44,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   41,   -1,   -1,   44,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   10,   41,   -1,   -1,   44,
   -1,   -1,   93,   -1,   -1,   -1,   -1,  260,  261,  262,
  263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
  273,  274,   -1,   -1,   10,   -1,   41,   -1,   -1,   44,
   -1,   -1,   -1,   -1,   93,  257,  258,  259,   -1,   -1,
  260,  261,  262,   -1,   -1,   -1,  268,  269,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   41,  278,  279,   44,  281,
  282,  283,  284,  285,  286,  287,  288,  289,  290,  291,
  292,  293,  294,  295,  296,  297,  298,  299,   93,   10,
  302,   -1,   -1,  260,  261,  262,  263,  264,  265,  266,
  267,  268,  269,  270,  271,  272,   -1,   -1,   -1,   -1,
   -1,   10,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,
   41,   10,   -1,   44,   -1,   -1,   -1,   -1,   -1,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,   41,   -1,   -1,   44,   -1,   -1,   -1,   -1,
   -1,   -1,   41,   -1,   -1,   44,   -1,  260,  261,  262,
  263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
   -1,   10,   93,   -1,   -1,   -1,   -1,   -1,   -1,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   41,   -1,   93,   44,   -1,   -1,   -1,   -1,
   41,  260,  261,  262,  263,  264,  265,  266,  267,  268,
  269,  270,  271,  272,   -1,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,   10,   41,
   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   93,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,   -1,   10,   -1,   -1,   41,
   -1,   -1,   44,   -1,   10,   -1,   -1,   41,   -1,   -1,
   44,   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,
  266,  267,  268,  269,   -1,   -1,   10,   41,   -1,   -1,
   44,   -1,   -1,   -1,   10,   41,   -1,   -1,   44,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,
   -1,   93,   -1,   -1,   10,   -1,   -1,   41,   -1,   93,
   44,   -1,   -1,   -1,   -1,   41,   -1,   -1,   44,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,   93,
   41,   -1,   -1,   44,   -1,   41,   -1,   93,   44,   41,
   -1,  260,  261,  262,  263,  264,  265,  266,  267,  268,
  269,  260,  261,  262,  263,  264,  265,  266,  267,   93,
   41,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   41,   -1,
   -1,   -1,   93,   -1,   -1,   -1,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   41,   -1,   -1,   -1,
   -1,  260,  261,  262,  263,  264,  265,  266,  267,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,   41,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,   41,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  260,  261,  262,  263,
  264,  265,  266,  267,   41,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,
  264,  265,  266,  267,  260,  261,  262,  263,  264,  265,
  266,  267,   41,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,
  264,  265,  266,  267,  260,  261,  262,  263,  264,  265,
  266,  267,   41,   -1,   -1,   -1,   -1,   -1,   -1,  260,
  261,  262,   -1,   -1,  260,  261,  262,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,   41,   -1,   -1,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,  260,  261,  262,
  263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
  273,  274,  275,  276,  277,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,   41,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,
  264,  265,  266,  267,  268,  269,  270,  271,  272,  273,
  274,  275,  276,  277,   41,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,   41,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,   41,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  260,  261,  262,  263,  264,  265,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,   41,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  260,  261,  262,  263,  264,  265,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,   41,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,   41,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   41,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   41,
   -1,   -1,  260,  261,  262,  263,  264,  265,  266,  267,
  268,  269,  270,  271,  272,  273,  274,  275,  276,  277,
   41,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,   41,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  260,  261,  262,  263,  264,  265,  266,  267,
  268,  269,  270,  271,  272,  273,  274,  275,  276,  277,
   41,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  260,  261,  262,  263,  264,  265,  266,  267,  268,  269,
  270,  271,  272,  273,  274,  275,  276,  277,   41,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,   41,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,   44,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  260,  261,  262,  263,  264,  265,  266,  267,
  268,  269,  270,  271,  272,  273,  274,  275,  276,  277,
  260,  261,  262,  263,  264,  265,  266,  267,  268,  269,
  270,  271,  272,  273,  274,  275,  276,  277,  260,  261,
  262,  263,  264,  265,  266,  267,  268,  269,  270,  271,
  272,  273,  274,  275,  276,  277,   44,   93,   -1,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,   44,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  260,  261,  262,  263,  264,  265,  266,  267,  268,  269,
  270,  271,  272,  273,  274,  275,  276,  277,   44,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,   44,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,  262,
  263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
  273,  274,  275,  276,  277,   44,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,   44,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  260,  261,  262,  263,  264,  265,  266,  267,
  268,  269,  270,  271,  272,  273,  274,  275,  276,  277,
   44,   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,
  266,  267,  268,  269,  270,  271,  272,  273,  274,  275,
  276,  277,   -1,   -1,   -1,  260,  261,  262,  263,  264,
  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  260,  261,  262,  263,  264,  265,  266,  267,
  268,  269,  270,  271,  272,  273,  274,  275,  276,  277,
   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,
  266,  267,  268,  269,  270,  271,  272,  273,  274,  275,
  276,  277,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  260,  261,  262,  263,  264,  265,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  260,  261,  262,  263,  264,  265,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  260,
  261,  262,  263,  264,  265,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  260,  261,  262,  263,
  264,  265,  266,  267,  268,  269,  270,  271,  272,  273,
  274,  275,  276,  277,  302,   -1,  304,   -1,  306,  307,
  308,  309,  310,  311,  312,  313,  314,  315,  316,  317,
  318,  319,  320,   -1,   -1,   -1,   -1,   -1,  326,  327,
  328,  329,  330,  331,   -1,  333,  334,  335,  336,  337,
  338,  339,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,  354,  355,  356,  357,
  358,  359,  360,  361,  362,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,  375,  376,  377,
  378,  379,  380,  381,  382,  383,  384,  385,  386,  387,
  388,  389,  390,  391,  392,  393,  287,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  301,  302,   -1,   -1,   -1,   -1,   -1,   -1,  309,   -1,
   -1,   -1,   -1,   -1,  315,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  323,  324,   -1,   -1,   -1,   -1,   -1,  330,
   -1,   -1,  333,  334,   -1,   -1,  337,  338,   -1,   -1,
   -1,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,
  260,  261,  262,  263,  264,  265,  266,  267,  268,  269,
  270,  271,  272,  273,  274,  275,  276,  277,
};
#define YYFINAL 1
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 415
#define YYUNDFTOKEN 557
#define YYTRANSLATE(a) ((a) > YYMAXTOKEN ? YYUNDFTOKEN : (a))
#if YYDEBUG
static const char *const yyname[] = {

"end-of-file",0,0,0,0,0,0,0,0,0,"'\\n'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,"'('","')'",0,0,"','",0,0,0,0,0,0,0,0,0,0,0,0,0,"':'",0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"'['",0,"']'",0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
"T_NUMBER","T_STRING","T_OP_LOGICNOT","T_OP_LOGICOR","T_OP_LOGICAND",
"T_OP_LOGICEQU","T_OP_LOGICGT","T_OP_LOGICLT","T_OP_LOGICGE","T_OP_LOGICLE",
"T_OP_LOGICNE","T_OP_ADD","T_OP_SUB","T_OP_OR","T_OP_XOR","T_OP_AND","T_OP_SHL",
"T_OP_SHR","T_OP_MUL","T_OP_DIV","T_OP_MOD","T_OP_NOT","T_OP_DEF","T_OP_BANK",
"T_OP_SIN","T_OP_COS","T_OP_TAN","T_OP_ASIN","T_OP_ACOS","T_OP_ATAN",
"T_OP_ATAN2","T_OP_FDIV","T_OP_FMUL","T_OP_ROUND","T_OP_CEIL","T_OP_FLOOR",
"T_OP_STRCMP","T_OP_STRIN","T_OP_STRSUB","T_OP_STRLEN","T_OP_STRCAT",
"T_OP_STRUPR","T_OP_STRLWR","NEG","T_LABEL","T_ID","T_POP_EQU","T_POP_SET",
"T_POP_EQUS","T_POP_INCLUDE","T_POP_PRINTF","T_POP_PRINTT","T_POP_PRINTV",
"T_POP_IF","T_POP_ELSE","T_POP_ENDC","T_POP_IMPORT","T_POP_EXPORT",
"T_POP_GLOBAL","T_POP_DB","T_POP_DS","T_POP_DW","T_POP_DL","T_POP_SECTION",
"T_POP_RB","T_POP_RW","T_POP_RL","T_POP_MACRO","T_POP_ENDM","T_POP_RSRESET",
"T_POP_RSSET","T_POP_INCBIN","T_POP_REPT","T_POP_CHARMAP","T_POP_SHIFT",
"T_POP_ENDR","T_POP_FAIL","T_POP_WARN","T_POP_PURGE","T_POP_POPS","T_POP_PUSHS",
"T_POP_POPO","T_POP_PUSHO","T_POP_OPT","T_SECT_WRAM0","T_SECT_VRAM",
"T_SECT_ROMX","T_SECT_ROM0","T_SECT_HRAM","T_SECT_WRAMX","T_SECT_SRAM",
"T_Z80_ADC","T_Z80_ADD","T_Z80_AND","T_Z80_BIT","T_Z80_CALL","T_Z80_CCF",
"T_Z80_CP","T_Z80_CPL","T_Z80_DAA","T_Z80_DEC","T_Z80_DI","T_Z80_EI","T_Z80_EX",
"T_Z80_HALT","T_Z80_INC","T_Z80_JP","T_Z80_JR","T_Z80_LD","T_Z80_LDI",
"T_Z80_LDD","T_Z80_LDIO","T_Z80_NOP","T_Z80_OR","T_Z80_POP","T_Z80_PUSH",
"T_Z80_RES","T_Z80_RET","T_Z80_RETI","T_Z80_RST","T_Z80_RL","T_Z80_RLA",
"T_Z80_RLC","T_Z80_RLCA","T_Z80_RR","T_Z80_RRA","T_Z80_RRC","T_Z80_RRCA",
"T_Z80_SBC","T_Z80_SCF","T_Z80_STOP","T_Z80_SLA","T_Z80_SRA","T_Z80_SRL",
"T_Z80_SUB","T_Z80_SWAP","T_Z80_XOR","T_MODE_A","T_MODE_B","T_MODE_C",
"T_MODE_C_IND","T_MODE_D","T_MODE_E","T_MODE_H","T_MODE_L","T_MODE_AF",
"T_MODE_BC","T_MODE_BC_IND","T_MODE_DE","T_MODE_DE_IND","T_MODE_SP",
"T_MODE_SP_IND","T_MODE_HL","T_MODE_HL_IND","T_MODE_HL_INDDEC",
"T_MODE_HL_INDINC","T_CC_NZ","T_CC_Z","T_CC_NC",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,"illegal-symbol",
};
static const char *const yyrule[] = {
"$accept : asmfile",
"asmfile : lines",
"lines :",
"lines : lines line '\\n'",
"line : label",
"line : label cpu_command",
"line : label macro",
"line : label simple_pseudoop",
"line : pseudoop",
"label :",
"label : T_LABEL",
"label : T_LABEL ':'",
"label : T_LABEL ':' ':'",
"$$1 :",
"macro : T_ID $$1 macroargs",
"macroargs :",
"macroargs : macroarg",
"macroargs : macroarg ',' macroargs",
"macroarg : T_STRING",
"pseudoop : equ",
"pseudoop : set",
"pseudoop : rb",
"pseudoop : rw",
"pseudoop : rl",
"pseudoop : equs",
"pseudoop : macrodef",
"simple_pseudoop : include",
"simple_pseudoop : printf",
"simple_pseudoop : printt",
"simple_pseudoop : printv",
"simple_pseudoop : if",
"simple_pseudoop : else",
"simple_pseudoop : endc",
"simple_pseudoop : import",
"simple_pseudoop : export",
"simple_pseudoop : global",
"simple_pseudoop : db",
"simple_pseudoop : dw",
"simple_pseudoop : dl",
"simple_pseudoop : ds",
"simple_pseudoop : section",
"simple_pseudoop : rsreset",
"simple_pseudoop : rsset",
"simple_pseudoop : incbin",
"simple_pseudoop : charmap",
"simple_pseudoop : rept",
"simple_pseudoop : shift",
"simple_pseudoop : fail",
"simple_pseudoop : warn",
"simple_pseudoop : purge",
"simple_pseudoop : pops",
"simple_pseudoop : pushs",
"simple_pseudoop : popo",
"simple_pseudoop : pusho",
"simple_pseudoop : opt",
"$$2 :",
"opt : T_POP_OPT $$2 opt_list",
"opt_list : opt_list_entry",
"opt_list : opt_list_entry ',' opt_list",
"opt_list_entry : T_STRING",
"popo : T_POP_POPO",
"pusho : T_POP_PUSHO",
"pops : T_POP_POPS",
"pushs : T_POP_PUSHS",
"fail : T_POP_FAIL string",
"warn : T_POP_WARN string",
"shift : T_POP_SHIFT",
"rept : T_POP_REPT const",
"macrodef : T_LABEL ':' T_POP_MACRO",
"equs : T_LABEL T_POP_EQUS string",
"rsset : T_POP_RSSET const",
"rsreset : T_POP_RSRESET",
"rl : T_LABEL T_POP_RL const",
"rw : T_LABEL T_POP_RW const",
"rb : T_LABEL T_POP_RB const",
"ds : T_POP_DS const",
"db : T_POP_DB constlist_8bit",
"dw : T_POP_DW constlist_16bit",
"dl : T_POP_DL constlist_32bit",
"$$3 :",
"purge : T_POP_PURGE $$3 purge_list",
"purge_list : purge_list_entry",
"purge_list : purge_list_entry ',' purge_list",
"purge_list_entry : T_ID",
"import : T_POP_IMPORT import_list",
"import_list : import_list_entry",
"import_list : import_list_entry ',' import_list",
"import_list_entry : T_ID",
"export : T_POP_EXPORT export_list",
"export_list : export_list_entry",
"export_list : export_list_entry ',' export_list",
"export_list_entry : T_ID",
"global : T_POP_GLOBAL global_list",
"global_list : global_list_entry",
"global_list : global_list_entry ',' global_list",
"global_list_entry : T_ID",
"equ : T_LABEL T_POP_EQU const",
"set : T_LABEL T_POP_SET const",
"include : T_POP_INCLUDE string",
"incbin : T_POP_INCBIN string",
"incbin : T_POP_INCBIN string ',' const ',' const",
"charmap : T_POP_CHARMAP string ',' string",
"charmap : T_POP_CHARMAP string ',' const",
"printt : T_POP_PRINTT string",
"printv : T_POP_PRINTV const",
"printf : T_POP_PRINTF const",
"if : T_POP_IF const",
"else : T_POP_ELSE",
"endc : T_POP_ENDC",
"const_3bit : const",
"constlist_8bit : constlist_8bit_entry",
"constlist_8bit : constlist_8bit_entry ',' constlist_8bit",
"constlist_8bit_entry :",
"constlist_8bit_entry : const_8bit",
"constlist_8bit_entry : string",
"constlist_16bit : constlist_16bit_entry",
"constlist_16bit : constlist_16bit_entry ',' constlist_16bit",
"constlist_16bit_entry :",
"constlist_16bit_entry : const_16bit",
"constlist_32bit : constlist_32bit_entry",
"constlist_32bit : constlist_32bit_entry ',' constlist_32bit",
"constlist_32bit_entry :",
"constlist_32bit_entry : relocconst",
"const_PCrel : relocconst",
"const_8bit : relocconst",
"const_16bit : relocconst",
"relocconst : T_ID",
"relocconst : T_NUMBER",
"relocconst : string",
"relocconst : T_OP_LOGICNOT relocconst",
"relocconst : relocconst T_OP_LOGICOR relocconst",
"relocconst : relocconst T_OP_LOGICAND relocconst",
"relocconst : relocconst T_OP_LOGICEQU relocconst",
"relocconst : relocconst T_OP_LOGICGT relocconst",
"relocconst : relocconst T_OP_LOGICLT relocconst",
"relocconst : relocconst T_OP_LOGICGE relocconst",
"relocconst : relocconst T_OP_LOGICLE relocconst",
"relocconst : relocconst T_OP_LOGICNE relocconst",
"relocconst : relocconst T_OP_ADD relocconst",
"relocconst : relocconst T_OP_SUB relocconst",
"relocconst : relocconst T_OP_XOR relocconst",
"relocconst : relocconst T_OP_OR relocconst",
"relocconst : relocconst T_OP_AND relocconst",
"relocconst : relocconst T_OP_SHL relocconst",
"relocconst : relocconst T_OP_SHR relocconst",
"relocconst : relocconst T_OP_MUL relocconst",
"relocconst : relocconst T_OP_DIV relocconst",
"relocconst : relocconst T_OP_MOD relocconst",
"relocconst : T_OP_ADD relocconst",
"relocconst : T_OP_SUB relocconst",
"relocconst : T_OP_NOT relocconst",
"relocconst : T_OP_BANK '(' T_ID ')'",
"$$4 :",
"relocconst : T_OP_DEF $$4 '(' T_ID ')'",
"relocconst : T_OP_ROUND '(' const ')'",
"relocconst : T_OP_CEIL '(' const ')'",
"relocconst : T_OP_FLOOR '(' const ')'",
"relocconst : T_OP_FDIV '(' const ',' const ')'",
"relocconst : T_OP_FMUL '(' const ',' const ')'",
"relocconst : T_OP_SIN '(' const ')'",
"relocconst : T_OP_COS '(' const ')'",
"relocconst : T_OP_TAN '(' const ')'",
"relocconst : T_OP_ASIN '(' const ')'",
"relocconst : T_OP_ACOS '(' const ')'",
"relocconst : T_OP_ATAN '(' const ')'",
"relocconst : T_OP_ATAN2 '(' const ',' const ')'",
"relocconst : T_OP_STRCMP '(' string ',' string ')'",
"relocconst : T_OP_STRIN '(' string ',' string ')'",
"relocconst : T_OP_STRLEN '(' string ')'",
"relocconst : '(' relocconst ')'",
"const : T_ID",
"const : T_NUMBER",
"const : string",
"const : T_OP_LOGICNOT const",
"const : const T_OP_LOGICOR const",
"const : const T_OP_LOGICAND const",
"const : const T_OP_LOGICEQU const",
"const : const T_OP_LOGICGT const",
"const : const T_OP_LOGICLT const",
"const : const T_OP_LOGICGE const",
"const : const T_OP_LOGICLE const",
"const : const T_OP_LOGICNE const",
"const : const T_OP_ADD const",
"const : const T_OP_SUB const",
"const : T_ID T_OP_SUB T_ID",
"const : const T_OP_XOR const",
"const : const T_OP_OR const",
"const : const T_OP_AND const",
"const : const T_OP_SHL const",
"const : const T_OP_SHR const",
"const : const T_OP_MUL const",
"const : const T_OP_DIV const",
"const : const T_OP_MOD const",
"const : T_OP_ADD const",
"const : T_OP_SUB const",
"const : T_OP_NOT const",
"const : T_OP_ROUND '(' const ')'",
"const : T_OP_CEIL '(' const ')'",
"const : T_OP_FLOOR '(' const ')'",
"const : T_OP_FDIV '(' const ',' const ')'",
"const : T_OP_FMUL '(' const ',' const ')'",
"const : T_OP_SIN '(' const ')'",
"const : T_OP_COS '(' const ')'",
"const : T_OP_TAN '(' const ')'",
"const : T_OP_ASIN '(' const ')'",
"const : T_OP_ACOS '(' const ')'",
"const : T_OP_ATAN '(' const ')'",
"const : T_OP_ATAN2 '(' const ',' const ')'",
"$$5 :",
"const : T_OP_DEF $$5 '(' T_ID ')'",
"const : T_OP_STRCMP '(' string ',' string ')'",
"const : T_OP_STRIN '(' string ',' string ')'",
"const : T_OP_STRLEN '(' string ')'",
"const : '(' const ')'",
"string : T_STRING",
"string : T_OP_STRSUB '(' string ',' const ',' const ')'",
"string : T_OP_STRCAT '(' string ',' string ')'",
"string : T_OP_STRUPR '(' string ')'",
"string : T_OP_STRLWR '(' string ')'",
"section : T_POP_SECTION string ',' sectiontype",
"section : T_POP_SECTION string ',' sectiontype '[' const ']'",
"section : T_POP_SECTION string ',' sectiontype ',' T_OP_BANK '[' const ']'",
"section : T_POP_SECTION string ',' sectiontype '[' const ']' ',' T_OP_BANK '[' const ']'",
"sectiontype : T_SECT_WRAM0",
"sectiontype : T_SECT_VRAM",
"sectiontype : T_SECT_ROMX",
"sectiontype : T_SECT_ROM0",
"sectiontype : T_SECT_HRAM",
"sectiontype : T_SECT_WRAMX",
"sectiontype : T_SECT_SRAM",
"cpu_command : z80_adc",
"cpu_command : z80_add",
"cpu_command : z80_and",
"cpu_command : z80_bit",
"cpu_command : z80_call",
"cpu_command : z80_ccf",
"cpu_command : z80_cp",
"cpu_command : z80_cpl",
"cpu_command : z80_daa",
"cpu_command : z80_dec",
"cpu_command : z80_di",
"cpu_command : z80_ei",
"cpu_command : z80_ex",
"cpu_command : z80_halt",
"cpu_command : z80_inc",
"cpu_command : z80_jp",
"cpu_command : z80_jr",
"cpu_command : z80_ld",
"cpu_command : z80_ldd",
"cpu_command : z80_ldi",
"cpu_command : z80_ldio",
"cpu_command : z80_nop",
"cpu_command : z80_or",
"cpu_command : z80_pop",
"cpu_command : z80_push",
"cpu_command : z80_res",
"cpu_command : z80_ret",
"cpu_command : z80_reti",
"cpu_command : z80_rl",
"cpu_command : z80_rla",
"cpu_command : z80_rlc",
"cpu_command : z80_rlca",
"cpu_command : z80_rr",
"cpu_command : z80_rra",
"cpu_command : z80_rrc",
"cpu_command : z80_rrca",
"cpu_command : z80_rst",
"cpu_command : z80_sbc",
"cpu_command : z80_scf",
"cpu_command : z80_set",
"cpu_command : z80_sla",
"cpu_command : z80_sra",
"cpu_command : z80_srl",
"cpu_command : z80_stop",
"cpu_command : z80_sub",
"cpu_command : z80_swap",
"cpu_command : z80_xor",
"z80_adc : T_Z80_ADC op_a_n",
"z80_adc : T_Z80_ADC op_a_r",
"z80_add : T_Z80_ADD op_a_n",
"z80_add : T_Z80_ADD op_a_r",
"z80_add : T_Z80_ADD op_hl_ss",
"z80_add : T_Z80_ADD T_MODE_SP comma const_8bit",
"z80_and : T_Z80_AND op_a_n",
"z80_and : T_Z80_AND op_a_r",
"z80_bit : T_Z80_BIT const_3bit comma reg_r",
"z80_call : T_Z80_CALL const_16bit",
"z80_call : T_Z80_CALL ccode comma const_16bit",
"z80_ccf : T_Z80_CCF",
"z80_cp : T_Z80_CP op_a_n",
"z80_cp : T_Z80_CP op_a_r",
"z80_cpl : T_Z80_CPL",
"z80_daa : T_Z80_DAA",
"z80_dec : T_Z80_DEC reg_r",
"z80_dec : T_Z80_DEC reg_ss",
"z80_di : T_Z80_DI",
"z80_ei : T_Z80_EI",
"z80_ex : T_Z80_EX T_MODE_HL comma T_MODE_SP_IND",
"z80_ex : T_Z80_EX T_MODE_SP_IND comma T_MODE_HL",
"z80_halt : T_Z80_HALT",
"z80_inc : T_Z80_INC reg_r",
"z80_inc : T_Z80_INC reg_ss",
"z80_jp : T_Z80_JP const_16bit",
"z80_jp : T_Z80_JP ccode comma const_16bit",
"z80_jp : T_Z80_JP T_MODE_HL_IND",
"z80_jp : T_Z80_JP T_MODE_HL",
"z80_jr : T_Z80_JR const_PCrel",
"z80_jr : T_Z80_JR ccode comma const_PCrel",
"z80_ldi : T_Z80_LDI T_MODE_HL_IND comma T_MODE_A",
"z80_ldi : T_Z80_LDI T_MODE_A comma T_MODE_HL",
"z80_ldi : T_Z80_LDI T_MODE_A comma T_MODE_HL_IND",
"z80_ldd : T_Z80_LDD T_MODE_HL_IND comma T_MODE_A",
"z80_ldd : T_Z80_LDD T_MODE_A comma T_MODE_HL",
"z80_ldd : T_Z80_LDD T_MODE_A comma T_MODE_HL_IND",
"z80_ldio : T_Z80_LDIO T_MODE_A comma op_mem_ind",
"z80_ldio : T_Z80_LDIO op_mem_ind comma T_MODE_A",
"z80_ld : z80_ld_mem",
"z80_ld : z80_ld_cind",
"z80_ld : z80_ld_rr",
"z80_ld : z80_ld_ss",
"z80_ld : z80_ld_hl",
"z80_ld : z80_ld_sp",
"z80_ld : z80_ld_r",
"z80_ld : z80_ld_a",
"z80_ld_hl : T_Z80_LD T_MODE_HL comma '[' T_MODE_SP const_8bit ']'",
"z80_ld_hl : T_Z80_LD T_MODE_HL comma T_MODE_SP const_8bit",
"z80_ld_hl : T_Z80_LD T_MODE_HL comma const_16bit",
"z80_ld_sp : T_Z80_LD T_MODE_SP comma T_MODE_HL",
"z80_ld_sp : T_Z80_LD T_MODE_SP comma const_16bit",
"z80_ld_mem : T_Z80_LD op_mem_ind comma T_MODE_SP",
"z80_ld_mem : T_Z80_LD op_mem_ind comma T_MODE_A",
"z80_ld_cind : T_Z80_LD T_MODE_C_IND comma T_MODE_A",
"z80_ld_rr : T_Z80_LD reg_rr comma T_MODE_A",
"z80_ld_r : T_Z80_LD reg_r comma const_8bit",
"z80_ld_r : T_Z80_LD reg_r comma reg_r",
"z80_ld_a : T_Z80_LD reg_r comma T_MODE_C_IND",
"z80_ld_a : T_Z80_LD reg_r comma reg_rr",
"z80_ld_a : T_Z80_LD reg_r comma op_mem_ind",
"z80_ld_ss : T_Z80_LD reg_ss comma const_16bit",
"z80_nop : T_Z80_NOP",
"z80_or : T_Z80_OR op_a_n",
"z80_or : T_Z80_OR op_a_r",
"z80_pop : T_Z80_POP reg_tt",
"z80_push : T_Z80_PUSH reg_tt",
"z80_res : T_Z80_RES const_3bit comma reg_r",
"z80_ret : T_Z80_RET",
"z80_ret : T_Z80_RET ccode",
"z80_reti : T_Z80_RETI",
"z80_rl : T_Z80_RL reg_r",
"z80_rla : T_Z80_RLA",
"z80_rlc : T_Z80_RLC reg_r",
"z80_rlca : T_Z80_RLCA",
"z80_rr : T_Z80_RR reg_r",
"z80_rra : T_Z80_RRA",
"z80_rrc : T_Z80_RRC reg_r",
"z80_rrca : T_Z80_RRCA",
"z80_rst : T_Z80_RST const_8bit",
"z80_sbc : T_Z80_SBC op_a_n",
"z80_sbc : T_Z80_SBC op_a_r",
"z80_scf : T_Z80_SCF",
"z80_set : T_POP_SET const_3bit comma reg_r",
"z80_sla : T_Z80_SLA reg_r",
"z80_sra : T_Z80_SRA reg_r",
"z80_srl : T_Z80_SRL reg_r",
"z80_stop : T_Z80_STOP",
"z80_sub : T_Z80_SUB op_a_n",
"z80_sub : T_Z80_SUB op_a_r",
"z80_swap : T_Z80_SWAP reg_r",
"z80_xor : T_Z80_XOR op_a_n",
"z80_xor : T_Z80_XOR op_a_r",
"op_mem_ind : '[' const_16bit ']'",
"op_hl_ss : reg_ss",
"op_hl_ss : T_MODE_HL comma reg_ss",
"op_a_r : reg_r",
"op_a_r : T_MODE_A comma reg_r",
"op_a_n : const_8bit",
"op_a_n : T_MODE_A comma const_8bit",
"comma : ','",
"ccode : T_CC_NZ",
"ccode : T_CC_Z",
"ccode : T_CC_NC",
"ccode : T_MODE_C",
"reg_r : T_MODE_B",
"reg_r : T_MODE_C",
"reg_r : T_MODE_D",
"reg_r : T_MODE_E",
"reg_r : T_MODE_H",
"reg_r : T_MODE_L",
"reg_r : T_MODE_HL_IND",
"reg_r : T_MODE_A",
"reg_tt : T_MODE_BC",
"reg_tt : T_MODE_DE",
"reg_tt : T_MODE_HL",
"reg_tt : T_MODE_AF",
"reg_ss : T_MODE_BC",
"reg_ss : T_MODE_DE",
"reg_ss : T_MODE_HL",
"reg_ss : T_MODE_SP",
"reg_rr : T_MODE_BC_IND",
"reg_rr : T_MODE_DE_IND",
"reg_rr : T_MODE_HL_INDINC",
"reg_rr : T_MODE_HL_INDDEC",

};
#endif

int      yydebug;
int      yynerrs;

int      yyerrflag;
int      yychar;
YYSTYPE  yyval;
YYSTYPE  yylval;

/* define the initial stack-sizes */
#ifdef YYSTACKSIZE
#undef YYMAXDEPTH
#define YYMAXDEPTH  YYSTACKSIZE
#else
#ifdef YYMAXDEPTH
#define YYSTACKSIZE YYMAXDEPTH
#else
#define YYSTACKSIZE 10000
#define YYMAXDEPTH  10000
#endif
#endif

#define YYINITSTACKSIZE 200

typedef struct {
    unsigned stacksize;
    YYINT    *s_base;
    YYINT    *s_mark;
    YYINT    *s_last;
    YYSTYPE  *l_base;
    YYSTYPE  *l_mark;
} YYSTACKDATA;
/* variables for the parser stack */
static YYSTACKDATA yystack;

#if YYDEBUG
#include <stdio.h>		/* needed for printf */
#endif

#include <stdlib.h>	/* needed for malloc, etc */
#include <string.h>	/* needed for memset */

/* allocate initial stack or double stack size, up to YYMAXDEPTH */
static int yygrowstack(YYSTACKDATA *data)
{
    int i;
    unsigned newsize;
    YYINT *newss;
    YYSTYPE *newvs;

    if ((newsize = data->stacksize) == 0)
        newsize = YYINITSTACKSIZE;
    else if (newsize >= YYMAXDEPTH)
        return YYENOMEM;
    else if ((newsize *= 2) > YYMAXDEPTH)
        newsize = YYMAXDEPTH;

    i = (int) (data->s_mark - data->s_base);
    newss = (YYINT *)realloc(data->s_base, newsize * sizeof(*newss));
    if (newss == 0)
        return YYENOMEM;

    data->s_base = newss;
    data->s_mark = newss + i;

    newvs = (YYSTYPE *)realloc(data->l_base, newsize * sizeof(*newvs));
    if (newvs == 0)
        return YYENOMEM;

    data->l_base = newvs;
    data->l_mark = newvs + i;

    data->stacksize = newsize;
    data->s_last = data->s_base + newsize - 1;
    return 0;
}

#if YYPURE || defined(YY_NO_LEAKS)
static void yyfreestack(YYSTACKDATA *data)
{
    free(data->s_base);
    free(data->l_base);
    memset(data, 0, sizeof(*data));
}
#else
#define yyfreestack(data) /* nothing */
#endif

#define YYABORT  goto yyabort
#define YYREJECT goto yyabort
#define YYACCEPT goto yyaccept
#define YYERROR  goto yyerrlab

int
YYPARSE_DECL()
{
    int yym, yyn, yystate;
#if YYDEBUG
    const char *yys;

    if ((yys = getenv("YYDEBUG")) != 0)
    {
        yyn = *yys;
        if (yyn >= '0' && yyn <= '9')
            yydebug = yyn - '0';
    }
#endif

    yynerrs = 0;
    yyerrflag = 0;
    yychar = YYEMPTY;
    yystate = 0;

#if YYPURE
    memset(&yystack, 0, sizeof(yystack));
#endif

    if (yystack.s_base == NULL && yygrowstack(&yystack) == YYENOMEM) goto yyoverflow;
    yystack.s_mark = yystack.s_base;
    yystack.l_mark = yystack.l_base;
    yystate = 0;
    *yystack.s_mark = 0;

yyloop:
    if ((yyn = yydefred[yystate]) != 0) goto yyreduce;
    if (yychar < 0)
    {
        if ((yychar = YYLEX) < 0) yychar = YYEOF;
#if YYDEBUG
        if (yydebug)
        {
            yys = yyname[YYTRANSLATE(yychar)];
            printf("%sdebug: state %d, reading %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
    }
    if ((yyn = yysindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: state %d, shifting to state %d\n",
                    YYPREFIX, yystate, yytable[yyn]);
#endif
        if (yystack.s_mark >= yystack.s_last && yygrowstack(&yystack) == YYENOMEM)
        {
            goto yyoverflow;
        }
        yystate = yytable[yyn];
        *++yystack.s_mark = yytable[yyn];
        *++yystack.l_mark = yylval;
        yychar = YYEMPTY;
        if (yyerrflag > 0)  --yyerrflag;
        goto yyloop;
    }
    if ((yyn = yyrindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
        yyn = yytable[yyn];
        goto yyreduce;
    }
    if (yyerrflag) goto yyinrecovery;

    YYERROR_CALL("syntax error");

    goto yyerrlab;

yyerrlab:
    ++yynerrs;

yyinrecovery:
    if (yyerrflag < 3)
    {
        yyerrflag = 3;
        for (;;)
        {
            if ((yyn = yysindex[*yystack.s_mark]) && (yyn += YYERRCODE) >= 0 &&
                    yyn <= YYTABLESIZE && yycheck[yyn] == YYERRCODE)
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: state %d, error recovery shifting\
 to state %d\n", YYPREFIX, *yystack.s_mark, yytable[yyn]);
#endif
                if (yystack.s_mark >= yystack.s_last && yygrowstack(&yystack) == YYENOMEM)
                {
                    goto yyoverflow;
                }
                yystate = yytable[yyn];
                *++yystack.s_mark = yytable[yyn];
                *++yystack.l_mark = yylval;
                goto yyloop;
            }
            else
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: error recovery discarding state %d\n",
                            YYPREFIX, *yystack.s_mark);
#endif
                if (yystack.s_mark <= yystack.s_base) goto yyabort;
                --yystack.s_mark;
                --yystack.l_mark;
            }
        }
    }
    else
    {
        if (yychar == YYEOF) goto yyabort;
#if YYDEBUG
        if (yydebug)
        {
            yys = yyname[YYTRANSLATE(yychar)];
            printf("%sdebug: state %d, error recovery discards token %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
        yychar = YYEMPTY;
        goto yyloop;
    }

yyreduce:
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: state %d, reducing by rule %d (%s)\n",
                YYPREFIX, yystate, yyn, yyrule[yyn]);
#endif
    yym = yylen[yyn];
    if (yym)
        yyval = yystack.l_mark[1-yym];
    else
        memset(&yyval, 0, sizeof yyval);
    switch (yyn)
    {
case 3:
#line 565 "src/asm/asmy.y"
	{
		nLineNo += 1;
		nTotalLines += 1;
	}
break;
case 10:
#line 577 "src/asm/asmy.y"
	{
		if (yystack.l_mark[0].tzSym[0] == '.')
			sym_AddLocalReloc(yystack.l_mark[0].tzSym);
		else
			sym_AddReloc(yystack.l_mark[0].tzSym);
	}
break;
case 11:
#line 582 "src/asm/asmy.y"
	{
		if (yystack.l_mark[-1].tzSym[0] == '.')
			sym_AddLocalReloc(yystack.l_mark[-1].tzSym);
		else
			sym_AddReloc(yystack.l_mark[-1].tzSym);
	}
break;
case 12:
#line 587 "src/asm/asmy.y"
	{
		sym_AddReloc(yystack.l_mark[-2].tzSym);
		sym_Export(yystack.l_mark[-2].tzSym);
	}
break;
case 13:
#line 592 "src/asm/asmy.y"
	{
		yy_set_state(LEX_STATE_MACROARGS);
	}
break;
case 14:
#line 594 "src/asm/asmy.y"
	{
		yy_set_state(LEX_STATE_NORMAL);

		if (!fstk_RunMacro(yystack.l_mark[-2].tzSym)) {
			fatalerror("Macro '%s' not defined", yystack.l_mark[-2].tzSym);
		}
	}
break;
case 18:
#line 606 "src/asm/asmy.y"
	{
		sym_AddNewMacroArg(yystack.l_mark[0].tzString);
	}
break;
case 55:
#line 648 "src/asm/asmy.y"
	{
		yy_set_state(LEX_STATE_MACROARGS);
	}
break;
case 56:
#line 650 "src/asm/asmy.y"
	{
		yy_set_state(LEX_STATE_NORMAL);
	}
break;
case 59:
#line 657 "src/asm/asmy.y"
	{
		opt_Parse(yystack.l_mark[0].tzString);
	}
break;
case 60:
#line 661 "src/asm/asmy.y"
	{
		opt_Pop();
	}
break;
case 61:
#line 665 "src/asm/asmy.y"
	{
		opt_Push();
	}
break;
case 62:
#line 669 "src/asm/asmy.y"
	{
		out_PopSection();
	}
break;
case 63:
#line 673 "src/asm/asmy.y"
	{
		out_PushSection();
	}
break;
case 64:
#line 677 "src/asm/asmy.y"
	{
		fatalerror("%s", yystack.l_mark[0].tzString);
	}
break;
case 65:
#line 681 "src/asm/asmy.y"
	{
		yyerror("%s", yystack.l_mark[0].tzString);
	}
break;
case 66:
#line 686 "src/asm/asmy.y"
	{ sym_ShiftCurrentMacroArgs(); }
break;
case 67:
#line 690 "src/asm/asmy.y"
	{
						copyrept();
						fstk_RunRept( yystack.l_mark[0].nConstValue );
					}
break;
case 68:
#line 697 "src/asm/asmy.y"
	{
						copymacro();
						sym_AddMacro(yystack.l_mark[-2].tzSym);
					}
break;
case 69:
#line 704 "src/asm/asmy.y"
	{ sym_AddString( yystack.l_mark[-2].tzSym, yystack.l_mark[0].tzString ); }
break;
case 70:
#line 708 "src/asm/asmy.y"
	{ sym_AddSet( "_RS", yystack.l_mark[0].nConstValue ); }
break;
case 71:
#line 712 "src/asm/asmy.y"
	{ sym_AddSet( "_RS", 0 ); }
break;
case 72:
#line 716 "src/asm/asmy.y"
	{
						sym_AddEqu( yystack.l_mark[-2].tzSym, sym_GetConstantValue("_RS") );
						sym_AddSet( "_RS", sym_GetConstantValue("_RS")+4*yystack.l_mark[0].nConstValue );
					}
break;
case 73:
#line 723 "src/asm/asmy.y"
	{
						sym_AddEqu( yystack.l_mark[-2].tzSym, sym_GetConstantValue("_RS") );
						sym_AddSet( "_RS", sym_GetConstantValue("_RS")+2*yystack.l_mark[0].nConstValue );
					}
break;
case 74:
#line 730 "src/asm/asmy.y"
	{
						sym_AddEqu( yystack.l_mark[-2].tzSym, sym_GetConstantValue("_RS") );
						sym_AddSet( "_RS", sym_GetConstantValue("_RS")+yystack.l_mark[0].nConstValue );
					}
break;
case 75:
#line 737 "src/asm/asmy.y"
	{ out_Skip( yystack.l_mark[0].nConstValue ); }
break;
case 79:
#line 750 "src/asm/asmy.y"
	{
						oDontExpandStrings = true;
					}
break;
case 80:
#line 754 "src/asm/asmy.y"
	{
						oDontExpandStrings = false;
					}
break;
case 83:
#line 763 "src/asm/asmy.y"
	{ sym_Purge(yystack.l_mark[0].tzSym); }
break;
case 87:
#line 773 "src/asm/asmy.y"
	{ sym_Import(yystack.l_mark[0].tzSym); }
break;
case 91:
#line 783 "src/asm/asmy.y"
	{ sym_Export(yystack.l_mark[0].tzSym); }
break;
case 95:
#line 793 "src/asm/asmy.y"
	{ sym_Global(yystack.l_mark[0].tzSym); }
break;
case 96:
#line 797 "src/asm/asmy.y"
	{ sym_AddEqu( yystack.l_mark[-2].tzSym, yystack.l_mark[0].nConstValue ); }
break;
case 97:
#line 801 "src/asm/asmy.y"
	{ sym_AddSet( yystack.l_mark[-2].tzSym, yystack.l_mark[0].nConstValue ); }
break;
case 98:
#line 805 "src/asm/asmy.y"
	{
						fstk_RunInclude(yystack.l_mark[0].tzString);
					}
break;
case 99:
#line 811 "src/asm/asmy.y"
	{ out_BinaryFile( yystack.l_mark[0].tzString ); }
break;
case 100:
#line 813 "src/asm/asmy.y"
	{
						out_BinaryFileSlice( yystack.l_mark[-4].tzString, yystack.l_mark[-2].nConstValue, yystack.l_mark[0].nConstValue );
					}
break;
case 101:
#line 819 "src/asm/asmy.y"
	{
						if(charmap_Add(yystack.l_mark[-2].tzString, yystack.l_mark[0].tzString[0] & 0xFF) == -1)
						{
							fprintf(stderr, "Error parsing charmap. Either you've added too many (%i), or the input character length is too long (%i)' : %s\n", MAXCHARMAPS, CHARMAPLENGTH, strerror(errno));
							yyerror("Error parsing charmap.");
						}
					}
break;
case 102:
#line 827 "src/asm/asmy.y"
	{
						if(charmap_Add(yystack.l_mark[-2].tzString, yystack.l_mark[0].nConstValue & 0xFF) == -1)
						{
							fprintf(stderr, "Error parsing charmap. Either you've added too many (%i), or the input character length is too long (%i)' : %s\n", MAXCHARMAPS, CHARMAPLENGTH, strerror(errno));
							yyerror("Error parsing charmap.");
						}
					}
break;
case 103:
#line 837 "src/asm/asmy.y"
	{
						if( nPass==1 )
							printf( "%s", yystack.l_mark[0].tzString );
					}
break;
case 104:
#line 844 "src/asm/asmy.y"
	{
						if( nPass==1 )
							printf( "$%lX", yystack.l_mark[0].nConstValue );
					}
break;
case 105:
#line 851 "src/asm/asmy.y"
	{
						if( nPass==1 )
							math_Print( yystack.l_mark[0].nConstValue );
					}
break;
case 106:
#line 858 "src/asm/asmy.y"
	{
						nIFDepth+=1;
						if( !yystack.l_mark[0].nConstValue )
						{
							if_skip_to_else();	/* will continue parsing just after ELSE or just at ENDC keyword */
						}
					}
break;
case 107:
#line 867 "src/asm/asmy.y"
	{
						if_skip_to_endc();		/* will continue parsing just at ENDC keyword */
					}
break;
case 108:
#line 873 "src/asm/asmy.y"
	{
						nIFDepth-=1;
					}
break;
case 109:
#line 879 "src/asm/asmy.y"
	{
						if( (yystack.l_mark[0].nConstValue<0) || (yystack.l_mark[0].nConstValue>7) )
						{
							yyerror("Immediate value must be 3-bit");
						}
						else
							yyval.nConstValue=yystack.l_mark[0].nConstValue&0x7;
					}
break;
case 112:
#line 893 "src/asm/asmy.y"
	{ out_Skip( 1 ); }
break;
case 113:
#line 894 "src/asm/asmy.y"
	{ out_RelByte( &yystack.l_mark[0].sVal ); }
break;
case 114:
#line 895 "src/asm/asmy.y"
	{ char *s; int length; s = yystack.l_mark[0].tzString; length = charmap_Convert(&s); out_AbsByteGroup(s, length); free(s); }
break;
case 117:
#line 902 "src/asm/asmy.y"
	{ out_Skip( 2 ); }
break;
case 118:
#line 903 "src/asm/asmy.y"
	{ out_RelWord( &yystack.l_mark[0].sVal ); }
break;
case 121:
#line 911 "src/asm/asmy.y"
	{ out_Skip( 4 ); }
break;
case 122:
#line 912 "src/asm/asmy.y"
	{ out_RelLong( &yystack.l_mark[0].sVal ); }
break;
case 123:
#line 917 "src/asm/asmy.y"
	{
						yyval.sVal = yystack.l_mark[0].sVal;
						if( !rpn_isPCRelative(&yystack.l_mark[0].sVal) )
							yyerror("Expression must be PC-relative");
					}
break;
case 124:
#line 925 "src/asm/asmy.y"
	{
						if( (!rpn_isReloc(&yystack.l_mark[0].sVal)) && ((yystack.l_mark[0].sVal.nVal<-128) || (yystack.l_mark[0].sVal.nVal>255)) )
						{
							yyerror("Expression must be 8-bit");
						}
						yyval.sVal=yystack.l_mark[0].sVal;
					}
break;
case 125:
#line 935 "src/asm/asmy.y"
	{
						if( (!rpn_isReloc(&yystack.l_mark[0].sVal)) && ((yystack.l_mark[0].sVal.nVal<-32768) || (yystack.l_mark[0].sVal.nVal>65535)) )
						{
							yyerror("Expression must be 16-bit");
						}
						yyval.sVal=yystack.l_mark[0].sVal;
					}
break;
case 126:
#line 946 "src/asm/asmy.y"
	{ rpn_Symbol(&yyval.sVal,yystack.l_mark[0].tzSym);	yyval.sVal.nVal = sym_GetValue(yystack.l_mark[0].tzSym); }
break;
case 127:
#line 948 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,yystack.l_mark[0].nConstValue);	yyval.sVal.nVal = yystack.l_mark[0].nConstValue; }
break;
case 128:
#line 950 "src/asm/asmy.y"
	{ char *s; int length; ULONG r; s = yystack.l_mark[0].tzString; length = charmap_Convert(&s); r = str2int2(s, length); free(s); rpn_Number(&yyval.sVal,r); yyval.sVal.nVal=r; }
break;
case 129:
#line 952 "src/asm/asmy.y"
	{ rpn_LOGNOT(&yyval.sVal,&yystack.l_mark[0].sVal); }
break;
case 130:
#line 954 "src/asm/asmy.y"
	{ rpn_LOGOR(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 131:
#line 956 "src/asm/asmy.y"
	{ rpn_LOGAND(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 132:
#line 958 "src/asm/asmy.y"
	{ rpn_LOGEQU(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 133:
#line 960 "src/asm/asmy.y"
	{ rpn_LOGGT(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 134:
#line 962 "src/asm/asmy.y"
	{ rpn_LOGLT(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 135:
#line 964 "src/asm/asmy.y"
	{ rpn_LOGGE(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 136:
#line 966 "src/asm/asmy.y"
	{ rpn_LOGLE(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 137:
#line 968 "src/asm/asmy.y"
	{ rpn_LOGNE(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 138:
#line 970 "src/asm/asmy.y"
	{ rpn_ADD(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 139:
#line 972 "src/asm/asmy.y"
	{ rpn_SUB(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 140:
#line 974 "src/asm/asmy.y"
	{ rpn_XOR(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 141:
#line 976 "src/asm/asmy.y"
	{ rpn_OR(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 142:
#line 978 "src/asm/asmy.y"
	{ rpn_AND(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 143:
#line 980 "src/asm/asmy.y"
	{ rpn_SHL(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 144:
#line 982 "src/asm/asmy.y"
	{ rpn_SHR(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 145:
#line 984 "src/asm/asmy.y"
	{ rpn_MUL(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 146:
#line 986 "src/asm/asmy.y"
	{ rpn_DIV(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 147:
#line 988 "src/asm/asmy.y"
	{ rpn_MOD(&yyval.sVal,&yystack.l_mark[-2].sVal,&yystack.l_mark[0].sVal); }
break;
case 148:
#line 990 "src/asm/asmy.y"
	{ yyval.sVal = yystack.l_mark[0].sVal; }
break;
case 149:
#line 992 "src/asm/asmy.y"
	{ rpn_UNNEG(&yyval.sVal,&yystack.l_mark[0].sVal); }
break;
case 150:
#line 994 "src/asm/asmy.y"
	{ rpn_UNNOT(&yyval.sVal,&yystack.l_mark[0].sVal); }
break;
case 151:
#line 996 "src/asm/asmy.y"
	{ rpn_Bank(&yyval.sVal,yystack.l_mark[-1].tzSym); yyval.sVal.nVal = 0; }
break;
case 152:
#line 997 "src/asm/asmy.y"
	{ oDontExpandStrings = true; }
break;
case 153:
#line 998 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,sym_isConstDefined(yystack.l_mark[-1].tzSym)); oDontExpandStrings = false; }
break;
case 154:
#line 999 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Round(yystack.l_mark[-1].nConstValue)); }
break;
case 155:
#line 1000 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Ceil(yystack.l_mark[-1].nConstValue)); }
break;
case 156:
#line 1001 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Floor(yystack.l_mark[-1].nConstValue)); }
break;
case 157:
#line 1002 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Div(yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue)); }
break;
case 158:
#line 1003 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Mul(yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue)); }
break;
case 159:
#line 1004 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Sin(yystack.l_mark[-1].nConstValue)); }
break;
case 160:
#line 1005 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Cos(yystack.l_mark[-1].nConstValue)); }
break;
case 161:
#line 1006 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_Tan(yystack.l_mark[-1].nConstValue)); }
break;
case 162:
#line 1007 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_ASin(yystack.l_mark[-1].nConstValue)); }
break;
case 163:
#line 1008 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_ACos(yystack.l_mark[-1].nConstValue)); }
break;
case 164:
#line 1009 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_ATan(yystack.l_mark[-1].nConstValue)); }
break;
case 165:
#line 1010 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,math_ATan2(yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue)); }
break;
case 166:
#line 1011 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,strcmp(yystack.l_mark[-3].tzString,yystack.l_mark[-1].tzString)); }
break;
case 167:
#line 1013 "src/asm/asmy.y"
	{
						char	*p;
  						if( (p=strstr(yystack.l_mark[-3].tzString,yystack.l_mark[-1].tzString))!=NULL )
						{
							rpn_Number(&yyval.sVal,p-yystack.l_mark[-3].tzString+1);
						}
						else
						{
							rpn_Number(&yyval.sVal,0);
						}
					}
break;
case 168:
#line 1024 "src/asm/asmy.y"
	{ rpn_Number(&yyval.sVal,strlen(yystack.l_mark[-1].tzString)); }
break;
case 169:
#line 1026 "src/asm/asmy.y"
	{ yyval.sVal = yystack.l_mark[-1].sVal; }
break;
case 170:
#line 1029 "src/asm/asmy.y"
	{ yyval.nConstValue = sym_GetConstantValue(yystack.l_mark[0].tzSym); }
break;
case 171:
#line 1030 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[0].nConstValue; }
break;
case 172:
#line 1031 "src/asm/asmy.y"
	{ yyval.nConstValue = str2int(yystack.l_mark[0].tzString); }
break;
case 173:
#line 1032 "src/asm/asmy.y"
	{ yyval.nConstValue = !yystack.l_mark[0].nConstValue; }
break;
case 174:
#line 1033 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue || yystack.l_mark[0].nConstValue; }
break;
case 175:
#line 1034 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue && yystack.l_mark[0].nConstValue; }
break;
case 176:
#line 1035 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue == yystack.l_mark[0].nConstValue; }
break;
case 177:
#line 1036 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue > yystack.l_mark[0].nConstValue; }
break;
case 178:
#line 1037 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue < yystack.l_mark[0].nConstValue; }
break;
case 179:
#line 1038 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue >= yystack.l_mark[0].nConstValue; }
break;
case 180:
#line 1039 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue <= yystack.l_mark[0].nConstValue; }
break;
case 181:
#line 1040 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue != yystack.l_mark[0].nConstValue; }
break;
case 182:
#line 1041 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue + yystack.l_mark[0].nConstValue; }
break;
case 183:
#line 1042 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue - yystack.l_mark[0].nConstValue; }
break;
case 184:
#line 1043 "src/asm/asmy.y"
	{ yyval.nConstValue = sym_GetDefinedValue(yystack.l_mark[-2].tzSym) - sym_GetDefinedValue(yystack.l_mark[0].tzSym); }
break;
case 185:
#line 1044 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue ^ yystack.l_mark[0].nConstValue; }
break;
case 186:
#line 1045 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue | yystack.l_mark[0].nConstValue; }
break;
case 187:
#line 1046 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue & yystack.l_mark[0].nConstValue; }
break;
case 188:
#line 1047 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue << yystack.l_mark[0].nConstValue; }
break;
case 189:
#line 1048 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue >> yystack.l_mark[0].nConstValue; }
break;
case 190:
#line 1049 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-2].nConstValue * yystack.l_mark[0].nConstValue; }
break;
case 191:
#line 1050 "src/asm/asmy.y"
	{
	if (yystack.l_mark[0].nConstValue == 0)
		fatalerror("division by zero");
	yyval.nConstValue = yystack.l_mark[-2].nConstValue / yystack.l_mark[0].nConstValue;
	}
break;
case 192:
#line 1055 "src/asm/asmy.y"
	{
	if (yystack.l_mark[0].nConstValue == 0)
		fatalerror("division by zero");
	yyval.nConstValue = yystack.l_mark[-2].nConstValue % yystack.l_mark[0].nConstValue;
	}
break;
case 193:
#line 1060 "src/asm/asmy.y"
	{ yyval.nConstValue = +yystack.l_mark[0].nConstValue; }
break;
case 194:
#line 1061 "src/asm/asmy.y"
	{ yyval.nConstValue = -yystack.l_mark[0].nConstValue; }
break;
case 195:
#line 1062 "src/asm/asmy.y"
	{ yyval.nConstValue = 0xFFFFFFFF^yystack.l_mark[0].nConstValue; }
break;
case 196:
#line 1063 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Round(yystack.l_mark[-1].nConstValue); }
break;
case 197:
#line 1064 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Ceil(yystack.l_mark[-1].nConstValue); }
break;
case 198:
#line 1065 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Floor(yystack.l_mark[-1].nConstValue); }
break;
case 199:
#line 1066 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Div(yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue); }
break;
case 200:
#line 1067 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Mul(yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue); }
break;
case 201:
#line 1068 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Sin(yystack.l_mark[-1].nConstValue); }
break;
case 202:
#line 1069 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Cos(yystack.l_mark[-1].nConstValue); }
break;
case 203:
#line 1070 "src/asm/asmy.y"
	{ yyval.nConstValue = math_Tan(yystack.l_mark[-1].nConstValue); }
break;
case 204:
#line 1071 "src/asm/asmy.y"
	{ yyval.nConstValue = math_ASin(yystack.l_mark[-1].nConstValue); }
break;
case 205:
#line 1072 "src/asm/asmy.y"
	{ yyval.nConstValue = math_ACos(yystack.l_mark[-1].nConstValue); }
break;
case 206:
#line 1073 "src/asm/asmy.y"
	{ yyval.nConstValue = math_ATan(yystack.l_mark[-1].nConstValue); }
break;
case 207:
#line 1074 "src/asm/asmy.y"
	{ yyval.nConstValue = math_ATan2(yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue); }
break;
case 208:
#line 1075 "src/asm/asmy.y"
	{ oDontExpandStrings = true; }
break;
case 209:
#line 1075 "src/asm/asmy.y"
	{ yyval.nConstValue = sym_isConstDefined(yystack.l_mark[-1].tzSym); oDontExpandStrings = false; }
break;
case 210:
#line 1076 "src/asm/asmy.y"
	{ yyval.nConstValue = strcmp( yystack.l_mark[-3].tzString, yystack.l_mark[-1].tzString ); }
break;
case 211:
#line 1078 "src/asm/asmy.y"
	{
						char	*p;
  						if( (p=strstr(yystack.l_mark[-3].tzString,yystack.l_mark[-1].tzString))!=NULL )
						{
							yyval.nConstValue = p-yystack.l_mark[-3].tzString+1;
						}
						else
						{
							yyval.nConstValue = 0;
						}
					}
break;
case 212:
#line 1089 "src/asm/asmy.y"
	{ yyval.nConstValue = strlen(yystack.l_mark[-1].tzString); }
break;
case 213:
#line 1090 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[-1].nConstValue; }
break;
case 214:
#line 1094 "src/asm/asmy.y"
	{ strcpy(yyval.tzString,yystack.l_mark[0].tzString); }
break;
case 215:
#line 1096 "src/asm/asmy.y"
	{ strncpy(yyval.tzString,yystack.l_mark[-5].tzString+yystack.l_mark[-3].nConstValue-1,yystack.l_mark[-1].nConstValue); yyval.tzString[yystack.l_mark[-1].nConstValue]=0; }
break;
case 216:
#line 1098 "src/asm/asmy.y"
	{ strcpy(yyval.tzString,yystack.l_mark[-3].tzString); strcat(yyval.tzString,yystack.l_mark[-1].tzString); }
break;
case 217:
#line 1100 "src/asm/asmy.y"
	{ strcpy(yyval.tzString,yystack.l_mark[-1].tzString); upperstring(yyval.tzString); }
break;
case 218:
#line 1102 "src/asm/asmy.y"
	{ strcpy(yyval.tzString,yystack.l_mark[-1].tzString); lowerstring(yyval.tzString); }
break;
case 219:
#line 1106 "src/asm/asmy.y"
	{
			out_NewSection(yystack.l_mark[-2].tzString,yystack.l_mark[0].nConstValue);
		}
break;
case 220:
#line 1110 "src/asm/asmy.y"
	{
			if( yystack.l_mark[-1].nConstValue>=0 && yystack.l_mark[-1].nConstValue<0x10000 )
				out_NewAbsSection(yystack.l_mark[-5].tzString,yystack.l_mark[-3].nConstValue,yystack.l_mark[-1].nConstValue,-1);
			else
				yyerror("Address $%x not 16-bit", yystack.l_mark[-1].nConstValue);
		}
break;
case 221:
#line 1117 "src/asm/asmy.y"
	{
			bankrangecheck(yystack.l_mark[-7].tzString, yystack.l_mark[-5].nConstValue, -1, yystack.l_mark[-1].nConstValue);
		}
break;
case 222:
#line 1121 "src/asm/asmy.y"
	{
			if (yystack.l_mark[-6].nConstValue < 0 || yystack.l_mark[-6].nConstValue > 0x10000) {
				yyerror("Address $%x not 16-bit", yystack.l_mark[-6].nConstValue);
			}
			bankrangecheck(yystack.l_mark[-10].tzString, yystack.l_mark[-8].nConstValue, yystack.l_mark[-6].nConstValue, yystack.l_mark[-1].nConstValue);
		}
break;
case 223:
#line 1130 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_WRAM0; }
break;
case 224:
#line 1131 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_VRAM; }
break;
case 225:
#line 1132 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_ROMX; }
break;
case 226:
#line 1133 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_ROM0; }
break;
case 227:
#line 1134 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_HRAM; }
break;
case 228:
#line 1135 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_WRAMX; }
break;
case 229:
#line 1136 "src/asm/asmy.y"
	{ yyval.nConstValue=SECT_SRAM; }
break;
case 277:
#line 1189 "src/asm/asmy.y"
	{ out_AbsByte(0xCE); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 278:
#line 1190 "src/asm/asmy.y"
	{ out_AbsByte(0x88|yystack.l_mark[0].nConstValue); }
break;
case 279:
#line 1193 "src/asm/asmy.y"
	{ out_AbsByte(0xC6); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 280:
#line 1194 "src/asm/asmy.y"
	{ out_AbsByte(0x80|yystack.l_mark[0].nConstValue); }
break;
case 281:
#line 1195 "src/asm/asmy.y"
	{ out_AbsByte(0x09|(yystack.l_mark[0].nConstValue<<4)); }
break;
case 282:
#line 1197 "src/asm/asmy.y"
	{ out_AbsByte(0xE8); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 283:
#line 1201 "src/asm/asmy.y"
	{ out_AbsByte(0xE6); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 284:
#line 1202 "src/asm/asmy.y"
	{ out_AbsByte(0xA0|yystack.l_mark[0].nConstValue); }
break;
case 285:
#line 1206 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x40|(yystack.l_mark[-2].nConstValue<<3)|yystack.l_mark[0].nConstValue); }
break;
case 286:
#line 1210 "src/asm/asmy.y"
	{ out_AbsByte(0xCD); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 287:
#line 1212 "src/asm/asmy.y"
	{ out_AbsByte(0xC4|(yystack.l_mark[-2].nConstValue<<3)); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 288:
#line 1216 "src/asm/asmy.y"
	{ out_AbsByte(0x3F); }
break;
case 289:
#line 1219 "src/asm/asmy.y"
	{ out_AbsByte(0xFE); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 290:
#line 1220 "src/asm/asmy.y"
	{ out_AbsByte(0xB8|yystack.l_mark[0].nConstValue); }
break;
case 291:
#line 1223 "src/asm/asmy.y"
	{ out_AbsByte(0x2F); }
break;
case 292:
#line 1226 "src/asm/asmy.y"
	{ out_AbsByte(0x27); }
break;
case 293:
#line 1230 "src/asm/asmy.y"
	{ out_AbsByte(0x05|(yystack.l_mark[0].nConstValue<<3)); }
break;
case 294:
#line 1232 "src/asm/asmy.y"
	{ out_AbsByte(0x0B|(yystack.l_mark[0].nConstValue<<4)); }
break;
case 295:
#line 1236 "src/asm/asmy.y"
	{ out_AbsByte(0xF3); }
break;
case 296:
#line 1240 "src/asm/asmy.y"
	{ out_AbsByte(0xFB); }
break;
case 297:
#line 1244 "src/asm/asmy.y"
	{ out_AbsByte(0xE3); }
break;
case 298:
#line 1246 "src/asm/asmy.y"
	{ out_AbsByte(0xE3); }
break;
case 299:
#line 1250 "src/asm/asmy.y"
	{
			out_AbsByte(0x76);
			if (CurrentOptions.haltnop) {
				out_AbsByte(0x00);
			}
		}
break;
case 300:
#line 1259 "src/asm/asmy.y"
	{ out_AbsByte(0x04|(yystack.l_mark[0].nConstValue<<3)); }
break;
case 301:
#line 1261 "src/asm/asmy.y"
	{ out_AbsByte(0x03|(yystack.l_mark[0].nConstValue<<4)); }
break;
case 302:
#line 1265 "src/asm/asmy.y"
	{ out_AbsByte(0xC3); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 303:
#line 1267 "src/asm/asmy.y"
	{ out_AbsByte(0xC2|(yystack.l_mark[-2].nConstValue<<3)); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 304:
#line 1269 "src/asm/asmy.y"
	{ out_AbsByte(0xE9); }
break;
case 305:
#line 1271 "src/asm/asmy.y"
	{ out_AbsByte(0xE9); }
break;
case 306:
#line 1275 "src/asm/asmy.y"
	{ out_AbsByte(0x18); out_PCRelByte(&yystack.l_mark[0].sVal); }
break;
case 307:
#line 1277 "src/asm/asmy.y"
	{ out_AbsByte(0x20|(yystack.l_mark[-2].nConstValue<<3)); out_PCRelByte(&yystack.l_mark[0].sVal); }
break;
case 308:
#line 1281 "src/asm/asmy.y"
	{ out_AbsByte(0x02|(2<<4)); }
break;
case 309:
#line 1283 "src/asm/asmy.y"
	{ out_AbsByte(0x0A|(2<<4)); }
break;
case 310:
#line 1285 "src/asm/asmy.y"
	{ out_AbsByte(0x0A|(2<<4)); }
break;
case 311:
#line 1289 "src/asm/asmy.y"
	{ out_AbsByte(0x02|(3<<4)); }
break;
case 312:
#line 1291 "src/asm/asmy.y"
	{ out_AbsByte(0x0A|(3<<4)); }
break;
case 313:
#line 1293 "src/asm/asmy.y"
	{ out_AbsByte(0x0A|(3<<4)); }
break;
case 314:
#line 1297 "src/asm/asmy.y"
	{
						rpn_CheckHRAM(&yystack.l_mark[0].sVal,&yystack.l_mark[0].sVal);

						if( (!rpn_isReloc(&yystack.l_mark[0].sVal))
						&&	(yystack.l_mark[0].sVal.nVal<0 || (yystack.l_mark[0].sVal.nVal>0xFF && yystack.l_mark[0].sVal.nVal<0xFF00) || yystack.l_mark[0].sVal.nVal>0xFFFF) )
						{
							yyerror("Source address $%x not in HRAM ($FF00 to $FFFE)", yystack.l_mark[0].sVal.nVal);
						}

						out_AbsByte(0xF0);
						yystack.l_mark[0].sVal.nVal&=0xFF;
						out_RelByte(&yystack.l_mark[0].sVal);
					}
break;
case 315:
#line 1311 "src/asm/asmy.y"
	{
						rpn_CheckHRAM(&yystack.l_mark[-2].sVal,&yystack.l_mark[-2].sVal);

						if( (!rpn_isReloc(&yystack.l_mark[-2].sVal))
						&&	(yystack.l_mark[-2].sVal.nVal<0 || (yystack.l_mark[-2].sVal.nVal>0xFF && yystack.l_mark[-2].sVal.nVal<0xFF00) || yystack.l_mark[-2].sVal.nVal>0xFFFF) )
						{
							yyerror("Destination address $%x not in HRAM ($FF00 to $FFFE)", yystack.l_mark[-2].sVal.nVal);
						}

						out_AbsByte(0xE0);
						yystack.l_mark[-2].sVal.nVal&=0xFF;
						out_RelByte(&yystack.l_mark[-2].sVal);
					}
break;
case 324:
#line 1337 "src/asm/asmy.y"
	{ out_AbsByte(0xF8); out_RelByte(&yystack.l_mark[-1].sVal); }
break;
case 325:
#line 1339 "src/asm/asmy.y"
	{ out_AbsByte(0xF8); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 326:
#line 1341 "src/asm/asmy.y"
	{ out_AbsByte(0x01|(REG_HL<<4)); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 327:
#line 1344 "src/asm/asmy.y"
	{ out_AbsByte(0xF9); }
break;
case 328:
#line 1346 "src/asm/asmy.y"
	{ out_AbsByte(0x01|(REG_SP<<4)); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 329:
#line 1350 "src/asm/asmy.y"
	{ out_AbsByte(0x08); out_RelWord(&yystack.l_mark[-2].sVal); }
break;
case 330:
#line 1352 "src/asm/asmy.y"
	{
						if( (!rpn_isReloc(&yystack.l_mark[-2].sVal)) && yystack.l_mark[-2].sVal.nVal>=0xFF00)
						{
							out_AbsByte(0xE0);
							out_AbsByte(yystack.l_mark[-2].sVal.nVal&0xFF);
						}
						else
						{
							out_AbsByte(0xEA);
							out_RelWord(&yystack.l_mark[-2].sVal);
						}
					}
break;
case 331:
#line 1367 "src/asm/asmy.y"
	{ out_AbsByte(0xE2); }
break;
case 332:
#line 1371 "src/asm/asmy.y"
	{ out_AbsByte(0x02|(yystack.l_mark[-2].nConstValue<<4)); }
break;
case 333:
#line 1375 "src/asm/asmy.y"
	{ out_AbsByte(0x06|(yystack.l_mark[-2].nConstValue<<3)); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 334:
#line 1377 "src/asm/asmy.y"
	{
						if( (yystack.l_mark[-2].nConstValue==REG_HL_IND) && (yystack.l_mark[0].nConstValue==REG_HL_IND) )
						{
							yyerror("LD [HL],[HL] not a valid instruction");
						}
						else
							out_AbsByte(0x40|(yystack.l_mark[-2].nConstValue<<3)|yystack.l_mark[0].nConstValue);
					}
break;
case 335:
#line 1388 "src/asm/asmy.y"
	{
						if( yystack.l_mark[-2].nConstValue==REG_A )
							out_AbsByte(0xF2);
						else
						{
							yyerror("Destination operand must be A");
						}
					}
break;
case 336:
#line 1397 "src/asm/asmy.y"
	{
						if( yystack.l_mark[-2].nConstValue==REG_A )
							out_AbsByte(0x0A|(yystack.l_mark[0].nConstValue<<4));
						else
						{
							yyerror("Destination operand must be A");
						}
					}
break;
case 337:
#line 1406 "src/asm/asmy.y"
	{
						if( yystack.l_mark[-2].nConstValue==REG_A )
						{
							if( (!rpn_isReloc(&yystack.l_mark[0].sVal)) && yystack.l_mark[0].sVal.nVal>=0xFF00 )
							{
								out_AbsByte(0xF0);
								out_AbsByte(yystack.l_mark[0].sVal.nVal&0xFF);
							}
							else
							{
								out_AbsByte(0xFA);
								out_RelWord(&yystack.l_mark[0].sVal);
							}
						}
						else
						{
							yyerror("Destination operand must be A");
						}
					}
break;
case 338:
#line 1428 "src/asm/asmy.y"
	{ out_AbsByte(0x01|(yystack.l_mark[-2].nConstValue<<4)); out_RelWord(&yystack.l_mark[0].sVal); }
break;
case 339:
#line 1432 "src/asm/asmy.y"
	{ out_AbsByte(0x00); }
break;
case 340:
#line 1436 "src/asm/asmy.y"
	{ out_AbsByte(0xF6); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 341:
#line 1438 "src/asm/asmy.y"
	{ out_AbsByte(0xB0|yystack.l_mark[0].nConstValue); }
break;
case 342:
#line 1442 "src/asm/asmy.y"
	{ out_AbsByte(0xC1|(yystack.l_mark[0].nConstValue<<4)); }
break;
case 343:
#line 1446 "src/asm/asmy.y"
	{ out_AbsByte(0xC5|(yystack.l_mark[0].nConstValue<<4)); }
break;
case 344:
#line 1450 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x80|(yystack.l_mark[-2].nConstValue<<3)|yystack.l_mark[0].nConstValue); }
break;
case 345:
#line 1454 "src/asm/asmy.y"
	{ out_AbsByte(0xC9); }
break;
case 346:
#line 1456 "src/asm/asmy.y"
	{ out_AbsByte(0xC0|(yystack.l_mark[0].nConstValue<<3)); }
break;
case 347:
#line 1460 "src/asm/asmy.y"
	{ out_AbsByte(0xD9); }
break;
case 348:
#line 1464 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x10|yystack.l_mark[0].nConstValue); }
break;
case 349:
#line 1468 "src/asm/asmy.y"
	{ out_AbsByte(0x17); }
break;
case 350:
#line 1472 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x00|yystack.l_mark[0].nConstValue); }
break;
case 351:
#line 1476 "src/asm/asmy.y"
	{ out_AbsByte(0x07); }
break;
case 352:
#line 1480 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x18|yystack.l_mark[0].nConstValue); }
break;
case 353:
#line 1484 "src/asm/asmy.y"
	{ out_AbsByte(0x1F); }
break;
case 354:
#line 1488 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x08|yystack.l_mark[0].nConstValue); }
break;
case 355:
#line 1492 "src/asm/asmy.y"
	{ out_AbsByte(0x0F); }
break;
case 356:
#line 1496 "src/asm/asmy.y"
	{
						if( rpn_isReloc(&yystack.l_mark[0].sVal) )
						{
							yyerror("Address for RST must be absolute");
						}
						else if( (yystack.l_mark[0].sVal.nVal&0x38)!=yystack.l_mark[0].sVal.nVal )
						{
							yyerror("Invalid address $%x for RST", yystack.l_mark[0].sVal.nVal);
						}
						else
							out_AbsByte(0xC7|yystack.l_mark[0].sVal.nVal);
					}
break;
case 357:
#line 1510 "src/asm/asmy.y"
	{ out_AbsByte(0xDE); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 358:
#line 1511 "src/asm/asmy.y"
	{ out_AbsByte(0x98|yystack.l_mark[0].nConstValue); }
break;
case 359:
#line 1515 "src/asm/asmy.y"
	{ out_AbsByte(0x37); }
break;
case 360:
#line 1519 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0xC0|(yystack.l_mark[-2].nConstValue<<3)|yystack.l_mark[0].nConstValue); }
break;
case 361:
#line 1523 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x20|yystack.l_mark[0].nConstValue); }
break;
case 362:
#line 1527 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x28|yystack.l_mark[0].nConstValue); }
break;
case 363:
#line 1531 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x38|yystack.l_mark[0].nConstValue); }
break;
case 364:
#line 1535 "src/asm/asmy.y"
	{ out_AbsByte(0x10); out_AbsByte(0x00); }
break;
case 365:
#line 1538 "src/asm/asmy.y"
	{ out_AbsByte(0xD6); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 366:
#line 1539 "src/asm/asmy.y"
	{ out_AbsByte(0x90|yystack.l_mark[0].nConstValue); }
break;
case 367:
#line 1543 "src/asm/asmy.y"
	{ out_AbsByte(0xCB); out_AbsByte(0x30|yystack.l_mark[0].nConstValue); }
break;
case 368:
#line 1546 "src/asm/asmy.y"
	{ out_AbsByte(0xEE); out_RelByte(&yystack.l_mark[0].sVal); }
break;
case 369:
#line 1547 "src/asm/asmy.y"
	{ out_AbsByte(0xA8|yystack.l_mark[0].nConstValue); }
break;
case 370:
#line 1550 "src/asm/asmy.y"
	{ yyval.sVal = yystack.l_mark[-1].sVal; }
break;
case 371:
#line 1553 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[0].nConstValue; }
break;
case 372:
#line 1554 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[0].nConstValue; }
break;
case 373:
#line 1557 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[0].nConstValue; }
break;
case 374:
#line 1558 "src/asm/asmy.y"
	{ yyval.nConstValue = yystack.l_mark[0].nConstValue; }
break;
case 375:
#line 1561 "src/asm/asmy.y"
	{ yyval.sVal = yystack.l_mark[0].sVal; }
break;
case 376:
#line 1562 "src/asm/asmy.y"
	{ yyval.sVal = yystack.l_mark[0].sVal; }
break;
case 378:
#line 1568 "src/asm/asmy.y"
	{ yyval.nConstValue = CC_NZ; }
break;
case 379:
#line 1569 "src/asm/asmy.y"
	{ yyval.nConstValue = CC_Z; }
break;
case 380:
#line 1570 "src/asm/asmy.y"
	{ yyval.nConstValue = CC_NC; }
break;
case 381:
#line 1571 "src/asm/asmy.y"
	{ yyval.nConstValue = CC_C; }
break;
case 382:
#line 1574 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_B; }
break;
case 383:
#line 1575 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_C; }
break;
case 384:
#line 1576 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_D; }
break;
case 385:
#line 1577 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_E; }
break;
case 386:
#line 1578 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_H; }
break;
case 387:
#line 1579 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_L; }
break;
case 388:
#line 1580 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_HL_IND; }
break;
case 389:
#line 1581 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_A; }
break;
case 390:
#line 1584 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_BC; }
break;
case 391:
#line 1585 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_DE; }
break;
case 392:
#line 1586 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_HL; }
break;
case 393:
#line 1587 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_AF; }
break;
case 394:
#line 1590 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_BC; }
break;
case 395:
#line 1591 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_DE; }
break;
case 396:
#line 1592 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_HL; }
break;
case 397:
#line 1593 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_SP; }
break;
case 398:
#line 1596 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_BC_IND; }
break;
case 399:
#line 1597 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_DE_IND; }
break;
case 400:
#line 1598 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_HL_INDINC; }
break;
case 401:
#line 1599 "src/asm/asmy.y"
	{ yyval.nConstValue = REG_HL_INDDEC; }
break;
#line 3740 "src/asm/asmy.c"
    }
    yystack.s_mark -= yym;
    yystate = *yystack.s_mark;
    yystack.l_mark -= yym;
    yym = yylhs[yyn];
    if (yystate == 0 && yym == 0)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: after reduction, shifting from state 0 to\
 state %d\n", YYPREFIX, YYFINAL);
#endif
        yystate = YYFINAL;
        *++yystack.s_mark = YYFINAL;
        *++yystack.l_mark = yyval;
        if (yychar < 0)
        {
            if ((yychar = YYLEX) < 0) yychar = YYEOF;
#if YYDEBUG
            if (yydebug)
            {
                yys = yyname[YYTRANSLATE(yychar)];
                printf("%sdebug: state %d, reading %d (%s)\n",
                        YYPREFIX, YYFINAL, yychar, yys);
            }
#endif
        }
        if (yychar == YYEOF) goto yyaccept;
        goto yyloop;
    }
    if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yystate)
        yystate = yytable[yyn];
    else
        yystate = yydgoto[yym];
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: after reduction, shifting from state %d \
to state %d\n", YYPREFIX, *yystack.s_mark, yystate);
#endif
    if (yystack.s_mark >= yystack.s_last && yygrowstack(&yystack) == YYENOMEM)
    {
        goto yyoverflow;
    }
    *++yystack.s_mark = (YYINT) yystate;
    *++yystack.l_mark = yyval;
    goto yyloop;

yyoverflow:
    YYERROR_CALL("yacc stack overflow");

yyabort:
    yyfreestack(&yystack);
    return (1);

yyaccept:
    yyfreestack(&yystack);
    return (0);
}
